# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging
import uuid
from collections import defaultdict
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.const import Chatstate
from nbxmpp.modules.security_labels import SecurityLabel

from gajim.common import app
from gajim.common import configpaths
from gajim.common import events
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.commands import CommandFailed
from gajim.common.const import AvatarSize
from gajim.common.const import Direction
from gajim.common.const import Draft
from gajim.common.const import SimpleClientState
from gajim.common.events import MessageSent
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.storage.archive import models as mod
from gajim.common.structs import ReplyData
from gajim.common.types import ChatContactT
from gajim.common.util.filesystem import write_file_async

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.chat_state_indicator import ChatStateIndicator
from gajim.gtk.emoji_chooser import EmojiChooser
from gajim.gtk.menus import get_encryption_menu
from gajim.gtk.menus import get_format_menu
from gajim.gtk.message_input import MessageInputTextView
from gajim.gtk.message_url_previews import MessageURLPreviews
from gajim.gtk.referenced_message import ReplyBox
from gajim.gtk.security_label_selector import SecurityLabelSelector
from gajim.gtk.timezone_hint import TimezoneHint
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import allow_send_message
from gajim.gtk.util.misc import is_message_correctable
from gajim.gtk.util.window import open_window
from gajim.gtk.voice_message_recorder_widget import VoiceMessageRecorderButton

log = logging.getLogger("gajim.gtk.messageactionsbox")


class MessageActionsBox(Gtk.Grid, EventHelper, SignalManager):
    @GObject.Property(type=bool, default=False)
    def compact(self) -> bool:  # pyright: ignore
        return self._compact

    @compact.setter
    def compact(self, compact: bool) -> None:
        self._set_compact(compact)

    def __init__(self) -> None:
        Gtk.Grid.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self._client: Client | None = None
        self._contact: ChatContactT | None = None
        # XEP-0308 Message Correction
        self._correcting: dict[ChatContactT, str | None] = defaultdict(lambda: None)
        self._last_message_id: dict[ChatContactT, str | None] = {}

        self._ui = get_builder("message_actions_box.ui")
        self.add_css_class("message-actions-box")

        self.attach(self._ui.box, 0, 0, 1, 1)

        self._compact = False

        self._ui.state_box_image.set_size_request(AvatarSize.CHAT, -1)
        self._ui.edit_box_image.set_size_request(AvatarSize.CHAT, -1)

        self._timezone_hint = TimezoneHint()
        self._ui.hint_box.append(self._timezone_hint)

        self._chat_state_indicator = ChatStateIndicator()
        self._ui.hint_box.append(self._chat_state_indicator)

        # For message replies
        self._reply_box = ReplyBox()
        self._ui.reply_box.append(self._reply_box)

        self._ui.send_message_button.set_visible(
            app.settings.get("show_send_message_button")
        )
        app.settings.bind_signal(
            "show_send_message_button", self._ui.send_message_button, "set_visible"
        )

        self._voice_recorder_button = VoiceMessageRecorderButton()
        self._ui.action_box.append(self._voice_recorder_button)
        self._ui.action_box.reorder_child_after(
            self._voice_recorder_button, self._ui.send_message_button
        )

        self._security_label_selector = SecurityLabelSelector()
        self._ui.action_box.append(self._security_label_selector)
        self._ui.action_box.reorder_child_after(
            self._security_label_selector, self._ui.input_wrapper
        )

        self._message_url_previews = MessageURLPreviews()
        self._ui.link_previews_box.append(self._message_url_previews)

        self._message_input = MessageInputTextView(self)
        self._connect(self._message_input, "buffer-changed", self._on_buffer_changed)
        self._connect(self._message_input, "paste-clipboard", self._on_paste_clipboard)

        self._ui.box.append(self._message_input.get_completion_popover())

        controller = Gtk.EventControllerKey()
        self._connect(controller, "key-pressed", self._on_message_input_key_pressed)
        self._message_input.add_controller(controller)

        self._ui.input_scrolled.set_child(self._message_input)

        self._wait_queue_resize = None
        self._vscrollbar_min_height = 0
        vadjustment = self._ui.input_scrolled.get_vadjustment()
        self._connect(vadjustment, "changed", self._on_input_scrolled_changed)
        self._connect(self._message_input, "realize", self._on_view_realize)

        self._ui.sendfile_button.set_tooltip_text(_("No File Transfer available"))
        self._ui.formattings_button.set_menu_model(get_format_menu())
        self._ui.encryption_menu_button.set_menu_model(get_encryption_menu())

        self._ui.emoticons_button.set_create_popup_func(self._on_emoji_create_popover)

        self._connect(
            self._ui.cancel_correction_button,
            "clicked",
            self._on_cancel_correction_clicked,
        )
        self._connect(
            self._ui.encryption_details_button,
            "clicked",
            self._on_encryption_details_clicked,
        )
        self._connect(
            self._ui.request_voice_button, "clicked", self._on_request_voice_clicked
        )

        app.plugin_manager.gui_extension_point(
            "message_actions_box", self, self._ui.action_box
        )

        # Watch the width of the action area itself and not the window's, the
        # room the buttons have also depends on the chat list and the
        # participants list being shown. Set up last, the breakpoint moves the
        # buttons around as soon as it applies.
        self._ui.box.remove(self._ui.action_box)
        breakpoint_bin = Adw.BreakpointBin(child=self._ui.action_box)
        # AdwBreakpointBin ignores the size of its child, it needs to be told
        # how far it may shrink. This is what the folded row needs.
        breakpoint_bin.set_size_request(320, 24)
        condition = Adw.BreakpointCondition.parse("max-width: 400sp")
        assert condition is not None
        compact_breakpoint = Adw.Breakpoint.new(condition)
        compact_breakpoint.add_setter(
            self,
            "compact",
            GObject.Value(GObject.TYPE_BOOLEAN, True),
        )
        breakpoint_bin.add_breakpoint(compact_breakpoint)
        self._ui.box.append(breakpoint_bin)

        self.register_events(
            [
                ("message-sent", ged.GUI2, self._on_message_sent),
                ("register-actions", ged.GUI2, self._on_register_actions),
                ("muc-disco-update", ged.GUI2, self._on_muc_disco_update),
                ("prepare-for-shutdown", ged.GUI2, self._on_prepare_for_shutdown),
            ]
        )

    def do_unroot(self) -> None:
        self._disconnect_all()
        self.unregister_events()

        if self._client is not None:
            self._client.disconnect_all_from_obj(self)
        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)

        self._security_label_selector.run_destroy()

        Gtk.Grid.do_unroot(self)
        app.check_finalize(self)

    def _on_register_actions(self, _event: events.RegisterActions) -> None:
        actions = [
            "quote",
            "reply",
            "correct-message",
        ]

        for action in actions:
            action = app.window.get_action(action)
            self._connect(action, "activate", self._on_action)

        action = app.window.get_action("set-encryption")
        self._connect(action, "change-state", self._change_encryption)

        # action = app.window.get_action('send-file-jingle')
        # self._connect(action, 'notify::enabled', self._on_send_file_enabled_changed)

        action = app.window.get_action("send-file-httpupload")
        self._connect(action, "notify::enabled", self._on_send_file_enabled_changed)

    def _on_muc_disco_update(self, event: events.MucDiscoUpdate) -> None:
        if self._contact is None or event.jid != self._contact.jid:
            return

        self._update_encryption_state()

    def _on_prepare_for_shutdown(self, event: events.PrepareForShutdown) -> None:
        self._store_draft()

    def get_current_contact(self) -> ChatContactT:
        assert self._contact is not None
        return self._contact

    def get_seclabel(self) -> SecurityLabel | None:
        return self._security_label_selector.get_seclabel()

    def get_message_input(self) -> MessageInputTextView:
        return self._message_input

    def _on_input_scrolled_changed(self, adj: Gtk.Adjustment) -> None:
        scrolled = self._ui.input_scrolled
        view = self._message_input

        """
        Fixes #12158. The textview would only show the last row(s),
        when entering a new row on some systems.
        """
        new_height = min(
            scrolled.get_max_content_height(),
            adj.get_upper() + view.get_margin_top() + view.get_margin_bottom(),
        )
        scrolled.set_size_request(-1, int(new_height))

        new_page_size = min(
            scrolled.get_size_request()[1]
            - (view.get_margin_top() + view.get_margin_bottom()),
            adj.get_upper(),
        )
        adj.set_page_size(new_page_size)

        """
        Hack for vscrollbar not requiring space and making textview higher
        TODO: doesn't resize immediately
        """
        scrolled.get_vscrollbar().set_visible(
            adj.get_upper()
            > scrolled.get_max_content_height() - 2 * self._vscrollbar_min_height
        )

        if self._contact is not None and self._wait_queue_resize is None:
            self._wait_queue_resize = GLib.timeout_add(100, self._queue_resize)

    def _queue_resize(self) -> bool:
        self._wait_queue_resize = None
        self._ui.input_scrolled.queue_resize()
        return GLib.SOURCE_REMOVE

    def _on_view_realize(self, _view: Gtk.TextView) -> None:
        min_size, _ = self._ui.input_scrolled.get_preferred_size()
        self._vscrollbar_min_height = min_size.height

    def _on_emoji_create_popover(self, button: Gtk.MenuButton) -> None:
        emoji_chooser = app.window.get_emoji_chooser()
        button.set_popover(emoji_chooser)
        emoji_chooser.set_emoji_picked_func(self._on_emoji_picked)

    def _on_emoji_picked(self, _emoji_chooser: EmojiChooser, text: str) -> None:
        self._message_input.insert_text(text)
        GLib.idle_add(self._message_input.grab_focus_delayed)

    def _on_action(
        self, action: Gio.SimpleAction, param: GLib.Variant | None
    ) -> int | None:
        if self._contact is None:
            return

        action_name = action.get_name()
        log.info("Activate action: %s", action_name)

        if action_name == "reply":
            assert param
            pk = param.get_uint32()
            original_message = app.storage.archive.get_message_with_pk(pk)
            if original_message is None:
                return
            self._enable_reply_mode(original_message)

        elif action_name == "correct-message":
            self.toggle_message_correction()

    def switch_contact(self, contact: ChatContactT) -> None:
        if self._client is not None:
            self._set_chatstate(False)
            self._client.disconnect_all_from_obj(self)

        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)

        self._client = app.get_client(contact.account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._store_draft()
        self._disable_reply_mode()

        self._contact = contact

        if isinstance(self._contact, GroupchatContact):
            self._contact.multi_connect(
                {
                    "state-changed": self._on_muc_state_changed,
                    "user-role-changed": self._on_muc_state_changed,
                }
            )
        elif isinstance(self._contact, GroupchatParticipant):
            self._contact.multi_connect(
                {
                    "user-joined": self._on_user_state_changed,
                    "user-left": self._on_user_state_changed,
                    "room-joined": self._on_user_state_changed,
                    "room-left": self._on_user_state_changed,
                }
            )

        self._update_encryption_state()
        self._update_send_file_button_tooltip()

        self._set_chatstate(True)

        self._message_url_previews.switch_contact(contact)
        self._message_input.switch_contact(contact)

        self._timezone_hint.switch_contact(contact)
        self._chat_state_indicator.switch_contact(contact)

        self._ui.edit_box.set_visible(self._is_correcting)
        self._restore_draft()

        self._security_label_selector.switch_contact(contact)

        self._update_message_input_state()

    def _store_draft(self) -> None:
        if self._contact is None:
            return

        text = self._message_input.get_text()

        reply_pk = None
        reply_data = self._reply_box.get_message_reply()
        if reply_data is not None:
            reply_pk = reply_data.pk

        draft = None
        if any([text, reply_pk]):
            draft = Draft(text, reply_pk)

        app.storage.archive.set_contact_value(
            self._contact.account, self._contact.jid, "draft", draft
        )
        self._contact.notify("draft-update", draft)

    def _restore_draft(self) -> None:
        assert self._contact is not None
        draft = app.storage.archive.get_contact_value(
            self._contact.account, self._contact.jid, "draft"
        )
        if draft is None:
            return

        self._message_input.insert_text(draft.text)

        if draft.reply_pk is None:
            return

        message = app.storage.archive.get_message_with_pk(draft.reply_pk)
        if message is None:
            return

        self._enable_reply_mode(message)

    def clear(self) -> None:
        if self._client is not None and not self._client.is_destroyed():
            self._set_chatstate(False)
            self._client.disconnect_all_from_obj(self)

        if self._contact is not None:
            self._store_draft()
            self._contact.disconnect_all_from_obj(self)

        self._security_label_selector.clear()

        if self._wait_queue_resize is not None:
            GLib.Source.remove(self._wait_queue_resize)
            self._wait_queue_resize = None

        self._contact = None
        self._client = None

    def get_text(self) -> str:
        return self._message_input.get_text()

    @property
    def _is_correcting(self) -> bool:
        if self._contact is None:
            return False

        return self._correcting[self._contact] is not None

    def _set_correcting(self, message_id: str | None) -> None:
        assert self._contact is not None
        self._correcting[self._contact] = message_id

    def _on_cancel_correction_clicked(self, _button: Gtk.Button) -> None:
        self._cancel_action()

    def insert_as_quote(self, text: str, *, clear: bool = False) -> None:
        if self._contact is None:
            return

        if clear:
            self._message_input.clear()
        self._message_input.insert_as_quote(text)

    def process_escape(self) -> bool:
        if not self._is_correcting and not self._reply_box.is_in_reply_mode:
            return False
        self._cancel_action()
        return True

    def _cancel_action(self) -> None:
        if self._is_correcting:
            self.toggle_message_correction()
        self._disable_reply_mode()

    def reset_state_after_send(self) -> None:
        self._message_input.clear()
        self._message_url_previews.clear()
        self._cancel_action()
        assert self._contact is not None
        app.storage.archive.set_contact_value(
            self._contact.account, self._contact.jid, "draft", None
        )
        self._contact.notify("draft-update", None)

    def toggle_message_correction(self) -> None:
        assert self._contact is not None

        if self._is_correcting:
            self._set_correcting(None)
            self._message_input.end_correction()
            self._ui.edit_box.set_visible(False)
            return

        last_message_id = self._last_message_id.get(self._contact)
        if last_message_id is None:
            return

        assert self._contact is not None
        message = app.storage.archive.get_message_with_id(
            self._contact.account,
            self._contact.jid,
            "message-id",
            last_message_id,
            default_options=True,
        )
        if message is None:
            return

        if not is_message_correctable(message):
            return

        self._set_correcting(last_message_id)

        message = message.get_last_correction() or message

        self._message_input.start_correction(message)
        self._ui.edit_box.set_visible(True)

        referenced_message = message.get_referenced_message()
        if referenced_message is not None:
            self._enable_reply_mode(referenced_message)

        if message.security_label is not None:
            self._security_label_selector.set_seclabel(
                message.security_label.label_hash
            )

    def get_last_message_id(self, contact: ChatContactT) -> str | None:
        return self._last_message_id.get(contact)

    def get_correction_id(self) -> str | None:
        assert self._contact is not None
        return self._correcting.get(self._contact)

    def _on_message_sent(self, event: MessageSent) -> None:
        message = event.message

        assert self._contact is not None

        if message.oob and message.oob[0].url == message.text:
            # Don't allow to correct HTTP Upload file transfer URLs
            self._last_message_id[self._contact] = None
        else:
            self._last_message_id[self._contact] = message.id

    def _on_client_state_changed(
        self, _client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        self._update_message_input_state()

    def _on_muc_state_changed(self, *args: Any) -> None:
        self._update_message_input_state()

    def _on_user_state_changed(self, *args: Any) -> None:
        self._update_message_input_state()

    def _update_message_input_state(self) -> None:
        assert self._client
        state = self._client.is_available()

        self._ui.visitor_menu_button.set_visible(False)

        if not state:
            self._ui.state_box_label.set_text(
                _("You are offline. Go online to send messages…")
            )
            self._ui.state_box_image.set_from_icon_name("lucide-cloud-off-symbolic")

        if isinstance(self._contact, GroupchatContact):
            state = self._contact.is_joined
            if self._contact.is_joined:
                self_contact = self._contact.get_self()
                assert self_contact
                state = not self_contact.role.is_visitor

                if self_contact.role.is_visitor:
                    self._ui.visitor_menu_button.set_visible(True)
                    self._ui.state_box_label.set_text(_("You are a visitor."))
                    self._ui.state_box_image.set_from_icon_name(
                        "lucide-mic-off-symbolic"
                    )

        if isinstance(self._contact, GroupchatParticipant):
            state = self._contact.is_available
            self._ui.state_box_label.set_text(
                _("You can’t send private messages to contacts if they are offline.")
            )
            self._ui.state_box_image.set_from_icon_name("lucide-cloud-off-symbolic")

        self._ui.state_box.set_visible(not state)

    def _set_chatstate(self, state: bool) -> None:
        assert self._client is not None
        if state:
            if self._message_input.has_text:
                self._client.get_module("Chatstate").set_chatstate(
                    self._contact, Chatstate.PAUSED
                )
            else:
                self._client.get_module("Chatstate").set_chatstate(
                    self._contact, Chatstate.ACTIVE
                )
        else:
            self._client.get_module("Chatstate").set_chatstate(
                self._contact, Chatstate.INACTIVE
            )

    def _change_encryption(self, action: Gio.SimpleAction, param: GLib.Variant) -> None:
        new_state = param.get_string()
        action_state = action.get_state()
        assert action_state is not None
        current_state = action_state.get_string()
        if current_state == new_state:
            return

        match new_state:
            case "" | "OMEMO" | "OpenPGP":
                pass

            case _:
                plugin = app.plugin_manager.encryption_plugins.get(new_state)
                if plugin is None:
                    # TODO: Add GUI error here
                    return

                if not plugin.activate_encryption(app.window.get_control()):
                    return

        contact = self.get_current_contact()
        contact.settings.set("encryption", new_state)
        action.set_state(GLib.Variant("s", new_state))

        self._update_encryption_button(new_state, is_available=True)
        self._update_encryption_details_button(new_state)

    def _update_encryption_state(self) -> None:
        contact = self.get_current_contact()
        encryption = contact.settings.get("encryption")
        encryption_available = self._is_encryption_available(contact)
        if not encryption_available and encryption:
            # Disable encryption if chat was encrypted before, but due to
            # changed circumstances, encryption is not applicable anymore
            # (i.e. group chat configuration changed).
            contact.settings.set("encryption", "")
            encryption = ""

        action = app.window.get_action("set-encryption")
        action.set_state(GLib.Variant("s", encryption))
        self._update_encryption_button(encryption, is_available=encryption_available)
        self._update_encryption_details_button(encryption)

    def _get_overflow_widgets(self) -> list[Gtk.Widget]:
        # The buttons next to the send button, in the order they appear in
        return [
            self._voice_recorder_button,
            self._ui.sendfile_button,
            self._ui.encryption_menu_button,
            self._ui.encryption_details_button,
        ]

    def _set_compact(self, compact: bool) -> None:
        if self._compact == compact:
            return
        self._compact = compact

        if compact:
            for widget in self._get_overflow_widgets():
                self._ui.action_box.remove(widget)
                self._ui.overflow_box.append(widget)
        else:
            after = self._ui.send_message_button
            for widget in self._get_overflow_widgets():
                self._ui.overflow_box.remove(widget)
                self._ui.action_box.insert_child_after(widget, after)
                after = widget

        self._ui.overflow_menu_button.set_visible(compact)

    def _update_encryption_button(self, encryption: str, *, is_available: bool) -> None:
        contact = self.get_current_contact()

        if is_available:
            tooltip = _("Choose encryption")

            if encryption in ("OMEMO", "OpenPGP", "PGP"):
                icon_name = "lucide-lock-symbolic"
            else:
                icon_name = "lucide-lock-open-symbolic"

        else:
            icon_name = "lucide-lock-open-symbolic"
            if isinstance(contact, GroupchatContact):
                tooltip = _("This is a public group chat. Encryption is not available.")
            elif isinstance(contact, GroupchatParticipant):
                tooltip = _("Encryption is not available in private chats")
            else:
                raise ValueError("Unexpected contact type: %s", type(contact))

        self._ui.encryption_menu_button.set_sensitive(is_available)
        self._ui.encryption_menu_button.set_tooltip_text(tooltip)
        self._ui.encryption_image.set_from_icon_name(icon_name)

    @staticmethod
    def _is_encryption_available(contact: ChatContactT) -> bool:
        if isinstance(contact, GroupchatContact):
            return contact.encryption_available
        elif isinstance(contact, GroupchatParticipant):
            return False
        return True

    def _update_encryption_details_button(self, encryption: str) -> None:
        encryption_state = {
            "visible": bool(encryption),
            "enc_type": encryption,
            "authenticated": False,
        }

        if encryption:
            if encryption in ("OMEMO", "OpenPGP"):
                encryption_state["authenticated"] = True
            else:
                # Only fire extension_point for plugins (i.e. not OMEMO)
                app.plugin_manager.extension_point(
                    f"encryption_state{encryption}",
                    app.window.get_control(),
                    encryption_state,
                )

        visible, enc_type, authenticated = encryption_state.values()
        assert isinstance(visible, bool)

        if authenticated:
            authenticated_string = _("and authenticated")
            icon_name = "lucide-shield-check-symbolic"
        else:
            authenticated_string = _("and NOT authenticated")
            icon_name = "lucide-shield-alert-symbolic"

        tooltip = _("%(type)s encryption is active %(authenticated)s.") % {
            "type": enc_type,
            "authenticated": authenticated_string,
        }

        if isinstance(self._contact, GroupchatContact) and visible:
            visible = self._contact.encryption_available

        self._ui.encryption_details_button.set_visible(visible)
        self._ui.encryption_details_button.set_tooltip_text(tooltip)
        self._ui.encryption_details_image.set_from_icon_name(icon_name)

    def _on_encryption_details_clicked(self, _button: Gtk.Button) -> None:
        contact = self.get_current_contact()
        encryption = contact.settings.get("encryption")
        if encryption in ("OMEMO", "OpenPGP"):
            if contact.is_groupchat:
                open_window(
                    "GroupchatDetails",
                    contact=contact,
                    page=f"encryption-{encryption.lower()}",
                )
                return

            if isinstance(contact, BareContact) and contact.is_self:
                window = open_window("Preferences")
                window.show_page(f"{contact.account}-encryption-{encryption.lower()}")
                return

            open_window(
                "ContactInfo",
                account=contact.account,
                contact=contact,
                page=f"encryption-{encryption.lower()}",
            )
            return

        app.plugin_manager.extension_point(
            f"encryption_dialog{encryption}", app.window.get_control()
        )

    def _on_send_file_enabled_changed(
        self, action: Gio.SimpleAction, _param: GObject.ParamSpec
    ) -> None:
        self._update_send_file_button_tooltip()

    def _update_send_file_button_tooltip(self):
        httpupload = app.window.get_action_enabled("send-file-httpupload")
        # jingle = app.window.get_action_enabled('send-file-jingle')
        jingle = False

        if not httpupload and not jingle:
            tooltip_text = _("No File Transfer available")
            self._ui.sendfile_button.set_tooltip_text(tooltip_text)
            return

        if self._contact is None:
            return

        client = app.get_client(self._contact.account)

        tooltip_text = _("Send File…")
        if httpupload and not jingle:
            max_file_size = client.get_module("HTTPUpload").max_file_size
            if max_file_size is not None:
                if app.settings.get("use_kib_mib"):
                    units = GLib.FormatSizeFlags.IEC_UNITS
                else:
                    units = GLib.FormatSizeFlags.DEFAULT
                max_file_size = GLib.format_size_full(int(max_file_size), units)
                tooltip_text = _("Send File (max. %s)…") % max_file_size

        self._ui.sendfile_button.set_tooltip_text(tooltip_text)

    def _on_buffer_changed(self, _message_input: MessageInputTextView) -> None:
        if self._contact is None:
            return

        has_text = self._message_input.has_text
        app.window.get_action("send-message").set_enabled(
            allow_send_message(has_text, self._contact)
        )

        self._message_url_previews.generate_url_previews(self._message_input.get_text())

        encryption_name = self._contact.settings.get("encryption")

        if has_text and encryption_name:
            app.plugin_manager.extension_point("typing" + encryption_name, self)

        client = app.get_client(self._contact.account)
        client.get_module("Chatstate").set_keyboard_activity(self._contact)
        if not has_text:
            client.get_module("Chatstate").set_chatstate_delayed(
                self._contact, Chatstate.ACTIVE
            )
            return

        client.get_module("Chatstate").set_chatstate(self._contact, Chatstate.COMPOSING)

    def _on_message_input_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        # pylint: disable=too-many-nested-blocks
        if state & Gdk.ModifierType.SHIFT_MASK:
            if state & Gdk.ModifierType.CONTROL_MASK:
                if keyval == Gdk.KEY_ISO_Left_Tab:
                    app.window.select_next_chat(Direction.PREV, unread_first=True)
                    return Gdk.EVENT_STOP

        if state & Gdk.ModifierType.CONTROL_MASK:
            if keyval == Gdk.KEY_Tab:
                app.window.select_next_chat(Direction.NEXT, unread_first=True)
                return Gdk.EVENT_STOP

            if keyval == Gdk.KEY_z:
                self._message_input.undo()
                return Gdk.EVENT_STOP

            if keyval == Gdk.KEY_y:
                self._message_input.redo()
                return Gdk.EVENT_STOP

            if keyval == Gdk.KEY_Up:
                self.toggle_message_correction()
                return Gdk.EVENT_STOP

        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            if state & Gdk.ModifierType.SHIFT_MASK:
                self._message_input.insert_newline()
                return Gdk.EVENT_STOP

            if state & Gdk.ModifierType.CONTROL_MASK:
                if not app.settings.get("send_on_ctrl_enter"):
                    self._message_input.insert_newline()
                    return Gdk.EVENT_STOP
            else:
                if app.settings.get("send_on_ctrl_enter"):
                    self._message_input.insert_newline()
                    return Gdk.EVENT_STOP

            assert self._contact is not None

            # Reset IMContext to clear preedit state
            self._message_input.reset_im_context()

            message = self._message_input.get_text()

            try:
                handled = app.commands.parse(self._contact.type_string, message)
            except CommandFailed:
                return Gdk.EVENT_STOP

            if handled:
                self._message_input.clear()
                return Gdk.EVENT_STOP

            if not app.account_is_available(self._contact.account):
                # we are not connected
                InformationAlertDialog(
                    _("You are offline"),
                    _("Your message can not be sent until you are connected."),
                )
                return Gdk.EVENT_STOP

            app.window.activate_action("win.send-message", None)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_request_voice_clicked(self, _button: Gtk.Button) -> None:
        self._ui.visitor_popover.popdown()
        app.window.activate_action("win.muc-request-voice", None)

    def _enable_reply_mode(self, original_message: mod.Message) -> None:
        assert self._contact is not None
        self._reply_box.enable_reply_mode(self._contact, original_message)
        self._message_input.grab_focus()

    def _disable_reply_mode(self, *args: Any) -> None:
        self._reply_box.disable_reply_mode()

    def get_message_reply(self) -> ReplyData | None:
        message_reply = self._reply_box.get_message_reply()
        if message_reply is None:
            return None

        return message_reply

    def get_open_graph_data(self) -> dict[str, OpenGraphData] | None:
        return self._message_url_previews.get_open_graph_data()

    def _on_paste_clipboard(self, textview: MessageInputTextView) -> None:
        clipboard = self.get_clipboard()
        formats = clipboard.get_formats()
        mime_types = formats.get_mime_types()

        log.debug(
            "Clipboard paste received, mime-types: %s, gtypes: %s",
            mime_types,
            formats.get_gtypes(),
        )

        if mime_types is None:
            log.warning("Could not determine clipboard mime types")
            return

        if "image/png" in mime_types or "image/bmp" in mime_types:
            self._ui.input_overlay.set_visible(True)
            self._ui.input_overlay_label.set_text(_("Processing image…"))
            textview.stop_emission_by_name("paste-clipboard")
            clipboard.read_texture_async(None, self._on_clipboard_read_texture_finished)
            return

        if "text/uri-list" in mime_types:
            # Prevent TextView from pasting the URIs as text:
            textview.stop_emission_by_name("paste-clipboard")
            clipboard.read_value_async(
                Gdk.FileList, 0, None, self._on_clipboard_read_value_finished
            )
            return

        log.info("Could not process clipboard mime type %s", mime_types)

    def _on_clipboard_read_value_finished(
        self,
        clipboard: Gdk.Clipboard,
        result: Gio.AsyncResult,
    ) -> None:
        try:
            file_list = clipboard.read_value_finish(result)
        except Exception as e:
            log.info("FileList read failed, trying text fallback: %s", e)
            clipboard.read_text_async(None, self._on_clipboard_text_fallback_finished)
            return

        if file_list is None or not isinstance(file_list, Gdk.FileList):
            log.info("Unexpected type received %r, trying text fallback", file_list)
            clipboard.read_text_async(None, self._on_clipboard_text_fallback_finished)
            return

        uris = [file.get_uri() for file in file_list.get_files()]
        app.window.activate_action("win.send-file", GLib.Variant("as", uris))

    def _on_clipboard_text_fallback_finished(
        self,
        clipboard: Gdk.Clipboard,
        result: Gio.AsyncResult,
    ) -> None:
        try:
            text = clipboard.read_text_finish(result)
        except Exception as e:
            log.warning("Text fallback paste also failed: %s", e)
            return

        if text:
            self._message_input.get_buffer().insert_at_cursor(text)

    def _on_clipboard_read_texture_finished(
        self,
        clipboard: Gdk.Clipboard,
        result: Gio.AsyncResult,
    ) -> None:
        try:
            texture = clipboard.read_texture_finish(result)
        except GLib.Error as e:
            self._ui.input_overlay.set_visible(False)
            InformationAlertDialog(_("Pasting Content Failed"), _("Error: %s") % e)
            return

        if texture is None:
            log.info("No image pasted")
            self._ui.input_overlay.set_visible(False)
            return

        temp_dir = configpaths.get_temp_dir()
        image_path = temp_dir / f"{uuid.uuid4()}.png"

        try:
            image = texture.save_to_png_bytes()
            image_bytes = image.get_data()
            if image_bytes is None:
                log.error("Could not process pasted image")
                self._ui.input_overlay.set_visible(False)
                return

        except GLib.Error as e:
            log.error("Error while trying to store pasted image: %s", e)
            self._ui.input_overlay.set_visible(False)
            return

        write_file_async(
            image_path, image_bytes, self._on_paste_image_saved, image_path
        )

    def _on_paste_image_saved(
        self, _successful: bool, error: GLib.Error | None, image_path: Path
    ) -> None:
        self._ui.input_overlay.set_visible(False)

        if error is not None:
            log.error("Error while trying to store pasted image: %s", error.message)
            return

        app.window.activate_action(
            "win.send-file", GLib.Variant("as", [image_path.as_uri()])
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Gtk

from gajim.gtk.util.misc import get_ui_string


@Gtk.Template.from_string(string=get_ui_string("preview/loading_box.ui"))
class LoadingBox(Gtk.Box):
    __gtype_name__ = "LoadingBox"



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import dataclasses
from enum import Enum

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.classes import SignalManager


class ChatTypeFilter(Enum):
    ALL = "all"
    CHAT = "chats"
    GROUPCHAT = "group_chats"


@dataclasses.dataclass
class ChatFilters:
    type: ChatTypeFilter = ChatTypeFilter.ALL
    group: str | None = None
    account: str | None = None


class ChatFilter(Gtk.Overlay, SignalManager):
    __gtype_name__ = "ChatFilter"

    __gsignals__ = {
        "filter-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self) -> None:
        Gtk.Overlay.__init__(self, halign=Gtk.Align.CENTER, tooltip_text=_("Filters"))
        SignalManager.__init__(self)

        self._block_signal = False

        menu_button = Gtk.MenuButton()
        self._connect(menu_button, "notify::active", self._on_menu_button_activated)
        icon = Gtk.Image.new_from_icon_name("lucide-list-filter-symbolic")
        menu_button.set_child(icon)
        self.set_child(menu_button)

        self._filter_active_dot = Gtk.Box(
            halign=Gtk.Align.END,
            valign=Gtk.Align.START,
            visible=False,
            width_request=12,
            height_request=12,
        )
        self._filter_active_dot.add_css_class("chat-filter-active")
        self.add_overlay(self._filter_active_dot)

        self._popover = Gtk.Popover(autohide=True, cascade_popdown=True)
        menu_button.set_popover(self._popover)

        popover_content = Gtk.Grid(row_spacing=6, column_spacing=12)
        popover_content.add_css_class("p-6")
        self._popover.set_child(popover_content)

        popover_title = Gtk.Label(label=_("Filter Chats"), margin_bottom=6)
        popover_title.add_css_class("bold")
        popover_content.attach(popover_title, 0, 0, 2, 1)

        chat_type_label = Gtk.Label(label=_("Chat Type"), halign=Gtk.Align.END)
        chat_type_label.add_css_class("dimmed")
        popover_content.attach(chat_type_label, 0, 1, 1, 1)

        chat_type_data = {
            ChatTypeFilter.ALL: _("All"),
            ChatTypeFilter.CHAT: _("Chats"),
            ChatTypeFilter.GROUPCHAT: _("Group Chats"),
        }
        self._chat_type_drop_down: GajimDropDown[ChatTypeFilter] = GajimDropDown(
            fixed_width=20, data=chat_type_data
        )
        self._connect(
            self._chat_type_drop_down, "notify::selected", self._on_chat_type_selected
        )
        popover_content.attach(self._chat_type_drop_down, 1, 1, 1, 1)

        roster_groups_label = Gtk.Label(label=_("Group"), halign=Gtk.Align.END)
        roster_groups_label.add_css_class("dimmed")
        popover_content.attach(roster_groups_label, 0, 2, 1, 1)

        self._roster_groups_drop_down: GajimDropDown[str | None] = GajimDropDown(
            fixed_width=20
        )
        self._roster_groups_drop_down.set_enable_search(True)
        self._connect(
            self._roster_groups_drop_down,
            "notify::selected",
            self._on_roster_group_selected,
        )
        popover_content.attach(self._roster_groups_drop_down, 1, 2, 1, 1)

        self._account_label = Gtk.Label(label=_("Account"), halign=Gtk.Align.END)
        self._account_label.add_css_class("dimmed")
        popover_content.attach(self._account_label, 0, 3, 1, 1)

        self._account_drop_down: GajimDropDown[str | None] = GajimDropDown(
            fixed_width=20
        )
        self._connect(
            self._account_drop_down,
            "notify::selected",
            self._on_account_selected,
        )
        popover_content.attach(self._account_drop_down, 1, 3, 1, 1)

        reset_button = Gtk.Button(
            label=_("Reset Filters"), halign=Gtk.Align.CENTER, margin_top=6
        )
        self._connect(reset_button, "clicked", self._on_reset_button_clicked)
        popover_content.attach(reset_button, 0, 4, 2, 1)

        self._populate_dropdowns()

    def do_unroot(self) -> None:
        Gtk.Overlay.do_unroot(self)
        self._disconnect_all()
        self._chat_type_drop_down.run_destroy()
        self._roster_groups_drop_down.run_destroy()
        self._account_drop_down.run_destroy()
        app.check_finalize(self)

    def get_filters(self) -> ChatFilters:
        chat_type_item = self._chat_type_drop_down.get_selected_key()
        assert chat_type_item is not None
        roster_group = self._roster_groups_drop_down.get_selected_key()
        account = self._account_drop_down.get_selected_key()

        return ChatFilters(
            type=chat_type_item,
            group=roster_group,
            account=account,
        )

    def set_filters(self, filters: ChatFilters) -> None:
        self._block_signal = True
        self._chat_type_drop_down.select_key(filters.type)
        self._roster_groups_drop_down.select_key(filters.group)
        self._account_drop_down.select_key(filters.account)
        self._block_signal = False
        self.emit("filter-changed")

    def reset(self) -> None:
        self._chat_type_drop_down.set_selected(0)
        self._roster_groups_drop_down.set_selected(0)
        self._account_drop_down.set_selected(0)

    def _on_menu_button_activated(
        self, menu_button: Gtk.MenuButton, *args: Any
    ) -> None:
        if not menu_button.get_active():
            return

        self._populate_dropdowns()

    def _populate_dropdowns(self) -> None:
        roster_group_key = self._roster_groups_drop_down.get_selected_key()
        account_key = self._account_drop_down.get_selected_key()

        roster_groups = self._get_roster_groups()
        accounts = self._get_accounts()

        show_account_filter = len(accounts) > 2
        self._account_drop_down.set_visible(show_account_filter)
        self._account_label.set_visible(show_account_filter)

        self._block_signal = True
        self._roster_groups_drop_down.set_data(roster_groups)
        self._roster_groups_drop_down.select_key(roster_group_key)

        self._account_drop_down.set_data(accounts)
        self._account_drop_down.select_key(account_key)
        self._block_signal = False

    def _update(self) -> None:
        filters = self.get_filters()
        active = (
            filters.type != ChatTypeFilter.ALL
            or filters.group is not None
            or filters.account is not None
        )
        if active:
            self.set_tooltip_text(_("Filters (active)"))
        else:
            self.set_tooltip_text(_("Filters"))

        self._filter_active_dot.set_visible(active)
        self._popover.grab_focus()

    def _on_chat_type_selected(
        self, _dropdown: GajimDropDown[ChatTypeFilter], *args: Any
    ) -> None:
        self._update()
        if not self._block_signal:
            self.emit("filter-changed")

    def _on_roster_group_selected(
        self, _dropdown: GajimDropDown[str | None], *args: Any
    ) -> None:
        self._update()
        if not self._block_signal:
            self.emit("filter-changed")

    def _on_account_selected(
        self, _dropdown: GajimDropDown[str | None], *args: Any
    ) -> None:
        self._update()
        if not self._block_signal:
            self.emit("filter-changed")

    def _on_reset_button_clicked(self, _button: Gtk.Button) -> None:
        self.reset()

    def _get_roster_groups(self) -> dict[str | None, str]:
        roster_groups: dict[str | None, str] = {None: _("All")}

        groups: set[str] = set()
        for account, _label in app.get_enabled_accounts_with_labels():
            client = app.get_client(account)
            groups.update(client.get_module("Roster").get_groups())

        for group in sorted(groups):
            roster_groups[group] = group

        return roster_groups

    def _get_accounts(self) -> dict[str | None, str]:
        accounts: dict[str | None, str] = {None: _("All")}
        for acc, label in app.get_enabled_accounts_with_labels():
            accounts[acc] = label
        return accounts



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Presence handler

from __future__ import annotations

from typing import Any
from typing import Literal

import time

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import idle
from gajim.common import types
from gajim.common.events import PresenceReceived
from gajim.common.events import ShowChanged
from gajim.common.events import SubscribedPresenceReceived
from gajim.common.events import SubscribePresenceReceived
from gajim.common.events import UnsubscribedPresenceReceived
from gajim.common.i18n import _
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import ResourceContact
from gajim.common.structs import PresenceData


class Presence(BaseModule):
    _nbxmpp_extends = "BasePresence"
    _nbxmpp_methods = [
        "subscribed",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._presence_received,
                typ="available",
                priority=50,
            ),
            StanzaHandler(
                name="presence",
                callback=self._presence_received,
                typ="unavailable",
                priority=50,
            ),
            StanzaHandler(
                name="presence",
                callback=self._subscribe_received,
                typ="subscribe",
                priority=49,
            ),
            StanzaHandler(
                name="presence",
                callback=self._subscribed_received,
                typ="subscribed",
                priority=49,
            ),
            StanzaHandler(
                name="presence",
                callback=self._unsubscribe_received,
                typ="unsubscribe",
                priority=49,
            ),
            StanzaHandler(
                name="presence",
                callback=self._unsubscribed_received,
                typ="unsubscribed",
                priority=49,
            ),
        ]

        self._presence_store: dict[JID, PresenceData] = {}

        # keep the jids we auto added (transports contacts) to not send the
        # SUBSCRIBED event to GUI
        self.automatically_added: set[JID] = set()

        # list of jid to auto-authorize
        self._jids_for_auto_auth: set[JID] = set()

    def _presence_received(
        self,
        _con: types.NBXMPPClient,
        stanza: nbxmpp.protocol.Presence,
        properties: PresenceProperties,
    ) -> None:

        self._log.info("Received from %s", properties.jid)

        if properties.from_muc:
            # MUC occupant presences are already handled in MUC module
            return

        assert properties.jid is not None
        muc = self._client.get_module("MUC").get_muc_data(properties.jid)
        if muc is not None:
            # Presence from the MUC itself, used for MUC avatar
            # handled in VCardAvatars module
            return

        roster_item = self._client.get_module("Roster").get_item(
            properties.jid.new_as_bare()
        )

        if roster_item is None and not properties.is_self_bare:
            # Handle only presence from roster contacts
            self._log.warning("Unknown presence received")
            self._log.warning(stanza)
            return

        presence_data = PresenceData.from_presence(properties)
        self._presence_store[properties.jid] = presence_data

        contact = self._client.get_module("Contacts").get_contact(properties.jid)
        assert isinstance(contact, BareContact | ResourceContact)
        contact.update_presence(presence_data)

        assert properties.type is not None
        assert properties.show is not None
        if properties.is_self_presence and not properties.type.is_unavailable:
            app.ged.raise_event(
                ShowChanged(account=self._account, show=properties.show.value)
            )
            return

        show = properties.show.value
        if properties.type.is_unavailable:
            show = "offline"

        event_attrs: dict[str, Any] = {
            "account": self._account,
            "stanza": stanza,
            "prio": properties.priority,
            "need_add_in_roster": False,
            "popup": False,
            "ptype": properties.type.value,
            "jid": properties.jid.new_as_bare(),
            "resource": properties.jid.resource,
            "id_": properties.id,
            "fjid": str(properties.jid),
            "timestamp": properties.timestamp,
            "avatar_sha": properties.avatar_sha,
            "user_nick": properties.nickname,
            "show": show,
            "new_show": show,
            "old_show": 0,
            "status": properties.status,
            "contact_list": [],
            "contact": None,
        }

        app.ged.raise_event(PresenceReceived(**event_attrs))

    def _subscribe_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: nbxmpp.protocol.Presence,
        properties: PresenceProperties,
    ) -> None:
        assert properties.jid is not None
        jid = properties.jid.new_as_bare()

        is_transport = app.jid_is_transport(str(properties.jid))
        auto_auth = app.settings.get_account_setting(self._account, "autoauth")

        self._log.info(
            "Received Subscribe: %s, transport: %s, auto_auth: %s, user_nick: %s",
            properties.jid,
            is_transport,
            auto_auth,
            properties.nickname,
        )

        if auto_auth or jid in self._jids_for_auto_auth:
            self._log.info("Auto respond with subscribed: %s", jid)
            self.subscribed(jid)
            return

        ignore_unknown_contacts = app.settings.get_account_setting(
            self._account, "ignore_unknown_contacts"
        )
        if ignore_unknown_contacts:
            self._log.info("Contact request ignored due to ignore_unknown_contacts")
            return

        status = properties.status or _("I would like to add you to my roster.")

        app.ged.raise_event(
            SubscribePresenceReceived(
                account=self._account,
                jid=jid,
                status=status,
                user_nick=properties.nickname,
                is_transport=is_transport,
            )
        )

        raise nbxmpp.NodeProcessed

    def _subscribed_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: nbxmpp.protocol.Presence,
        properties: PresenceProperties,
    ) -> None:
        assert properties.jid is not None
        self._log.info("Received Subscribed: %s", properties.jid)
        bare_jid = properties.jid.new_as_bare()
        if bare_jid in self.automatically_added:
            self.automatically_added.discard(bare_jid)
            raise nbxmpp.NodeProcessed

        app.ged.raise_event(
            SubscribedPresenceReceived(account=self._account, jid=properties.jid)
        )
        raise nbxmpp.NodeProcessed

    def _unsubscribe_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: nbxmpp.protocol.Presence,
        properties: PresenceProperties,
    ) -> None:
        self._log.info("Received Unsubscribe: %s", properties.jid)
        raise nbxmpp.NodeProcessed

    def _unsubscribed_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: nbxmpp.protocol.Presence,
        properties: PresenceProperties,
    ) -> None:
        assert properties.jid is not None
        self._log.info("Received Unsubscribed: %s", properties.jid)
        app.ged.raise_event(
            UnsubscribedPresenceReceived(
                account=self._account, jid=properties.jid.new_as_bare()
            )
        )
        raise nbxmpp.NodeProcessed

    def unsubscribed(self, jid: JID) -> None:
        self._log.info("Unsubscribed: %s", jid)
        self._jids_for_auto_auth.discard(jid)
        self._nbxmpp("BasePresence").unsubscribed(jid)

    def unsubscribe(self, jid: JID) -> None:
        self._log.info("Unsubscribe from %s", jid)
        self._jids_for_auto_auth.discard(jid)
        self._nbxmpp("BasePresence").unsubscribe(jid)

    def subscribe(
        self,
        jid: JID,
        msg: str | None = None,
        name: str | None = None,
        groups: list[str] | set[str] | None = None,
        auto_auth: bool = False,
    ) -> None:
        self._log.info("Request Subscription to %s", jid)

        if auto_auth:
            self._jids_for_auto_auth.add(jid)

        self._client.get_module("Roster").set_item(jid, name, groups=groups)
        self._nbxmpp("BasePresence").subscribe(
            jid, status=msg, nick=self._client.get_own_contact().name
        )

    def get_presence(
        self,
        to: str | JID | None = None,
        typ: Literal["probe", "unavailable"] | None = None,
        show: str | None = None,
        status: str | None = None,
        nick: str | None = None,
        caps: bool = True,
        idle_time: bool = False,
    ) -> nbxmpp.Presence:
        if show not in ("chat", "away", "xa", "dnd"):
            # Gajim sometimes passes invalid show values here
            # until this is fixed this is a workaround
            show = None
        presence = nbxmpp.Presence(to=to, typ=typ, show=show, status=status)
        if nick is not None:
            nick_tag = presence.setTag("nick", namespace=Namespace.NICK)
            nick_tag.setData(nick)

        if idle_time and app.is_installed("IDLE") and app.settings.get("autoaway"):
            idle_sec = idle.Monitor.get_idle_sec()
            time_ = time.strftime(
                "%Y-%m-%dT%H:%M:%SZ", time.gmtime(time.time() - idle_sec)
            )

            idle_node = presence.setTag("idle", namespace=Namespace.IDLE)
            idle_node.setAttr("since", time_)

        own_entity_caps = self._client.get_module("Caps").get_own_caps()
        if own_entity_caps is not None and typ != "unavailable":
            presence.setTag(
                "c", namespace=Namespace.CAPS, attrs=own_entity_caps._asdict()
            )

        return presence

    def send_presence(self, *args: Any, **kwargs: Any) -> None:
        if not app.account_is_connected(self._account):
            return
        presence = self.get_presence(*args, **kwargs)
        app.plugin_manager.extension_point("send-presence", self._account, presence)
        self._log.debug("Send presence:\n%s", presence)
        self._client.send_stanza(presence)



# Copyright (C) 2006 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006-2007 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Lukas Petrovicky <lukas AT petrovicky.net>
#                    Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging
import sys

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common import configpaths
from gajim.common import events
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import SimpleClientState
from gajim.common.dbus.menu import DBusMenu
from gajim.common.dbus.menu import DBusMenuItem
from gajim.common.dbus.statusnotifieritem import StatusNotifierItemService
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.util.status import get_global_show
from gajim.common.util.status import get_uf_show

from gajim.gtk.util.icons import get_tray_icon_name
from gajim.gtk.util.window import open_window

log = logging.getLogger("gajim.gtk.trayicon")

if sys.platform == "win32" or TYPE_CHECKING:
    import pystray
    from PIL import Image


class TrayIcon:
    def __init__(self) -> None:
        app.settings.connect_signal("show_trayicon", self._on_setting_changed)

        if sys.platform == "win32":
            self._backend = WindowsTrayIcon()
        elif sys.platform == "darwin":
            self._backend = NoneBackend()
        else:
            self._backend = LinuxTrayIcon()

    def _on_setting_changed(self, value: bool, *args: Any) -> None:
        self._backend.set_enabled(value)

    def connect_unread_widget(self, widget: Gtk.Widget, signal: str) -> None:
        widget.connect(signal, self._on_unread_count_changed)

    def _on_unread_count_changed(self, *args: Any) -> None:
        if not app.settings.get("trayicon_notification_on_events"):
            return
        self._backend.update_state()

    def is_visible(self) -> bool:
        return self._backend.is_visible()

    def shutdown(self) -> None:
        self._backend.shutdown()


class NoneBackend:
    def update_state(self, init: bool = False) -> None:
        pass

    def set_enabled(self, enabled: bool) -> None:
        pass

    def is_visible(self) -> bool:
        return False

    def shutdown(self) -> None:
        pass


class TrayIconBackend(EventHelper):
    def __init__(self) -> None:
        EventHelper.__init__(self)

        self.register_events(
            [
                ("our-show", ged.GUI1, self._on_our_show),
                ("account-enabled", ged.GUI1, self._on_account_enabled),
            ]
        )

        for client in app.get_clients():
            client.connect_signal("state-changed", self._on_client_state_changed)

    def update_state(self, init: bool = False) -> None:
        raise NotImplementedError

    def is_visible(self) -> bool:
        raise NotImplementedError

    def set_enabled(self, enabled: bool) -> None:
        raise NotImplementedError

    def _on_account_enabled(self, event: events.AccountEnabled) -> None:
        client = app.get_client(event.account)
        client.connect_signal("state-changed", self._on_client_state_changed)

    def _on_client_state_changed(
        self, _client: Client, _signal_name: str, _state: SimpleClientState
    ) -> None:
        self.update_state()

    def _on_our_show(self, _event: events.ShowChanged) -> None:
        self.update_state()

    @staticmethod
    def _on_start_chat() -> None:
        app.app.activate_action("start-chat", GLib.Variant("as", ["", ""]))

    @staticmethod
    def _on_sounds_mute() -> None:
        current_state = app.settings.get("sounds_on")
        app.settings.set("sounds_on", not current_state)

    def _on_show_hide(self) -> None:
        if not app.window.is_visible() or app.window.is_suspended():
            app.window.show_window()
            app.window.present()
        else:
            app.window.hide_window()

    @staticmethod
    def _on_preferences() -> None:
        open_window("Preferences")

    @staticmethod
    def _on_quit() -> None:
        app.app.start_shutdown()

    @staticmethod
    def _on_status_changed(status: str) -> None:
        app.app.change_status(status=status)

    def _on_activate(self) -> None:
        self._on_show_hide()


class WindowsTrayIcon(TrayIconBackend):
    def __init__(self) -> None:
        TrayIconBackend.__init__(self)
        self._tray_icon = self._create_tray_icon()

        if app.settings.get("show_trayicon"):
            self._tray_icon.run_detached()

    def update_state(self, init: bool = False) -> None:
        if not init and app.window.get_total_unread_count():
            self._tray_icon.icon = self._get_icon("message-new")
            return

        show = get_global_show()
        self._tray_icon.icon = self._get_icon(show)

    def set_enabled(self, enabled: bool) -> None:
        self._tray_icon.stop()

        if enabled:
            self._tray_icon = self._create_tray_icon()
            self._tray_icon.run_detached()

    def is_visible(self) -> bool:
        return self._tray_icon.visible

    def shutdown(self) -> None:
        self._tray_icon.stop()

    def _create_tray_icon(self) -> pystray.Icon:
        menu_items: list[pystray.MenuItem] = [
            pystray.MenuItem(
                text=_("Show/Hide Window"),
                action=lambda: GLib.idle_add(self._on_show_hide),
                default=True,
            ),
            pystray.MenuItem(
                _("Status"),
                pystray.Menu(
                    pystray.MenuItem(
                        text=get_uf_show("online"),
                        action=lambda: GLib.idle_add(self._on_status_changed, "online"),
                    ),
                    pystray.MenuItem(
                        text=get_uf_show("away"),
                        action=lambda: GLib.idle_add(self._on_status_changed, "away"),
                    ),
                    pystray.MenuItem(
                        text=get_uf_show("xa"),
                        action=lambda: GLib.idle_add(self._on_status_changed, "xa"),
                    ),
                    pystray.MenuItem(
                        text=get_uf_show("dnd"),
                        action=lambda: GLib.idle_add(self._on_status_changed, "dnd"),
                    ),
                    pystray.MenuItem(
                        text=get_uf_show("offline"),
                        action=lambda: GLib.idle_add(
                            self._on_status_changed, "offline"
                        ),
                    ),
                ),
            ),
            pystray.MenuItem(
                text=_("Start Chat…"),
                action=lambda: GLib.idle_add(self._on_start_chat),
            ),
            pystray.MenuItem(
                text=_("Mute Sounds"),
                action=lambda: GLib.idle_add(self._on_sounds_mute),
                checked=self._get_sound_toggle_state,
            ),
            pystray.MenuItem(
                text=_("Preferences"),
                action=lambda: GLib.idle_add(self._on_preferences),
            ),
            pystray.Menu.SEPARATOR,
            pystray.MenuItem(
                text=_("Quit"),
                action=lambda: GLib.idle_add(self._on_quit),
            ),
        ]

        return pystray.Icon(
            "Gajim", icon=self._get_icon("online"), menu=pystray.Menu(*menu_items)
        )

    @staticmethod
    def _get_sound_toggle_state(_item_name: str) -> bool:
        return not app.settings.get("sounds_on")

    @staticmethod
    def _get_icon(icon_name: str) -> Image.Image:
        tray_icon_name = get_tray_icon_name(icon_name)
        path = (
            configpaths.get("ICONS")
            / "hicolor"
            / "32x32"
            / "status"
            / f"{tray_icon_name}.png"
        )
        with Image.open(path) as image:
            image.load()
        return image


class LinuxTrayIcon(TrayIconBackend):
    def __init__(self) -> None:
        TrayIconBackend.__init__(self)

        self._shutdown = False
        self._tray_icon: StatusNotifierItemService | None = None

        self._theme_path = None
        if not app.is_flatpak():
            self._theme_path = str(configpaths.get("ICONS"))

        Gio.bus_get(Gio.BusType.SESSION, callback=self._on_finish)

    def _on_finish(self, obj: Any, res: Gio.AsyncResult) -> None:
        try:
            connection = Gio.bus_get_finish(res)
        except Exception as error:
            log.error(error)
            return

        self._tray_icon = StatusNotifierItemService(
            connection, self._get_menu(), self._theme_path, self._on_activate
        )

        log.info("Tray icon init successful")

        if app.settings.get("show_trayicon"):
            self._tray_icon.register()
            self.update_state(init=True)

    def _get_menu(self) -> DBusMenu:
        toogle_state = int(not app.settings.get("sounds_on"))
        return DBusMenu(
            items=[
                DBusMenuItem(
                    id=1, label=_("Show/Hide Window"), callback=self._on_show_hide
                ),
                DBusMenuItem(
                    id=2,
                    label=_("Status"),
                    children_display="submenu",
                    children=[
                        DBusMenuItem(
                            id=3,
                            label=get_uf_show("online"),
                            callback=lambda: self._on_status_changed("online"),
                        ),
                        DBusMenuItem(
                            id=4,
                            label=get_uf_show("away"),
                            callback=lambda: self._on_status_changed("away"),
                        ),
                        DBusMenuItem(
                            id=5,
                            label=get_uf_show("xa"),
                            callback=lambda: self._on_status_changed("xa"),
                        ),
                        DBusMenuItem(
                            id=6,
                            label=get_uf_show("dnd"),
                            callback=lambda: self._on_status_changed("dnd"),
                        ),
                        DBusMenuItem(id=7, type="separator"),
                        DBusMenuItem(
                            id=8,
                            label=get_uf_show("offline"),
                            callback=lambda: self._on_status_changed("offline"),
                        ),
                    ],
                ),
                DBusMenuItem(
                    id=9, label=_("Start Chat…"), callback=self._on_start_chat
                ),
                DBusMenuItem(
                    id=10,
                    label=_("Mute Sounds"),
                    toggle_type="checkmark",
                    toggle_state=toogle_state,
                    callback=self._on_sounds_mute,
                ),
                DBusMenuItem(
                    id=11, label=_("Preferences"), callback=self._on_preferences
                ),
                DBusMenuItem(id=12, type="separator"),
                DBusMenuItem(id=13, label=_("Quit"), callback=self._on_quit),
            ]
        )

    def update_state(self, init: bool = False) -> None:
        if self._tray_icon is None:
            # Tray icon is not initialized
            return

        if self._shutdown:
            # Shutdown in progress, don't update icon
            return

        if not init and app.window.get_total_unread_count():
            icon_name = get_tray_icon_name("message-new")
            self._tray_icon.set_icon(icon_name)
            return

        show = get_global_show()
        icon_name = get_tray_icon_name(show)
        self._tray_icon.set_icon(icon_name)

    def set_enabled(self, enabled: bool) -> None:
        if self._tray_icon is None:
            return

        log.info("Set tray icon enabled: %s", enabled)

        if enabled:
            self._tray_icon.register()
            self.update_state(init=True)
        else:
            self._tray_icon.unregister()

    def is_visible(self) -> bool:
        if self._tray_icon is None:
            return False
        return self._tray_icon.get_visible()

    def shutdown(self) -> None:
        if self._tray_icon is None:
            return
        self._shutdown = True
        self._tray_icon.unregister()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import sys
import uuid

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import configpaths
from gajim.common import events
from gajim.common import ged
from gajim.common.commands import ChatCommands
from gajim.common.const import CallType
from gajim.common.const import Display
from gajim.common.ged import EventHelper
from gajim.common.helpers import idle_add_once
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Message
from gajim.common.structs import OutgoingMessage
from gajim.common.types import ChatContactT
from gajim.common.util.muc import message_needs_highlight
from gajim.common.util.preview import get_preview_data
from gajim.common.util.text import remove_invalid_xml_chars

from gajim.gtk.activity_page import ActivityPage
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.chat_banner import ChatBanner
from gajim.gtk.chat_function_page import ChatFunctionPage
from gajim.gtk.chat_function_page import FunctionMode
from gajim.gtk.control import ChatControl
from gajim.gtk.message_actions_box import MessageActionsBox
from gajim.gtk.message_input import MessageInputTextView
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import allow_send_message
from gajim.gtk.util.misc import gi_gui_package_version
from gajim.gtk.util.misc import transform_to_inverted_bool
from gajim.gtk.util.window import open_window

log = logging.getLogger("gajim.gtk.chatstack")


class ChatStack(Gtk.Stack, EventHelper, SignalManager):
    __gtype_name__ = "ChatStack"

    def __init__(self):
        Gtk.Stack.__init__(self, hexpand=True, vexpand=True)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self.set_hhomogeneous(False)
        self.set_vhomogeneous(False)

        self._current_contact: ChatContactT | None = None
        self._last_quoted_id: int | None = None

        self.add_named(ChatPlaceholderBox(), "empty")

        self._activity_page = ActivityPage()
        self.add_named(self._activity_page, "activity")

        self._chat_function_page = ChatFunctionPage()
        self._chat_function_page.connect("finish", self._on_function_finished)
        self._chat_function_page.connect("message", self._on_function_message)
        self.add_named(self._chat_function_page, "function")

        self._chat_banner = ChatBanner()
        self._chat_control = ChatControl()
        self._message_action_box = MessageActionsBox()

        app.commands.connect("command-error", self._on_command_signal)
        app.commands.connect("command-not-found", self._on_command_signal)
        app.commands.connect("command-result", self._on_command_signal)

        self._banner_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._banner_box.append(self._chat_banner)
        self._banner_box.append(Gtk.Separator(margin_start=12, margin_end=12))

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.append(self._banner_box)
        box.append(self._chat_control.get_widget())
        box.append(self._message_action_box)

        dnd_icon = Gtk.Image.new_from_icon_name("lucide-paperclip-symbolic")
        dnd_icon.set_vexpand(True)
        dnd_icon.set_valign(Gtk.Align.END)
        dnd_icon.set_pixel_size(64)

        dnd_label = Gtk.Label(label=_("Drop files here"))
        dnd_label.set_max_width_chars(40)
        dnd_label.set_vexpand(True)
        dnd_label.set_valign(Gtk.Align.START)
        dnd_label.add_css_class("title-3")

        self._drop_area = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=18)
        self._drop_area.set_visible(False)
        self._drop_area.set_hexpand(True)
        self._drop_area.set_vexpand(True)
        self._drop_area.add_css_class("background")
        self._drop_area.append(dnd_icon)
        self._drop_area.append(dnd_label)

        overlay = Gtk.Overlay()
        overlay.add_overlay(self._drop_area)
        overlay.set_child(box)

        # TODO: support dnd for contacts (MUC invitations)

        format_builder = Gdk.ContentFormatsBuilder()
        format_builder.add_gtype(Gdk.Texture)
        format_builder.add_gtype(Gdk.FileList)

        drop_target = Gtk.DropTarget(
            actions=Gdk.DragAction.COPY, formats=format_builder.to_formats()
        )
        self._connect(drop_target, "accept", self._on_drop_accept)
        self._connect(drop_target, "drop", self._on_file_drop)
        self._connect(drop_target, "enter", self._on_drag_enter)
        self._connect(drop_target, "leave", self._on_drag_leave)
        overlay.add_controller(drop_target)

        controller = Gtk.EventControllerKey()
        self._connect(controller, "key-pressed", self._on_key_pressed)
        self.add_controller(controller)

        self.add_named(overlay, "controls")

        self.register_events(
            [
                ("message-received", 85, self._on_message_received),
                ("muc-disco-update", 85, self._on_muc_disco_update),
                ("account-connected", 85, self._on_account_state),
                ("account-disconnected", 85, self._on_account_state),
                ("register-actions", ged.GUI1, self._on_register_actions),
                ("feature-discovered", ged.GUI1, self._on_feature_discovered),
            ]
        )

    def do_unroot(self) -> None:
        Gtk.Stack.do_unroot(self)
        self._disconnect_all()
        self.unregister_events()
        app.check_finalize(self)

    def _on_register_actions(self, _event: events.RegisterActions) -> None:
        actions = [
            "send-file",
            "send-file-httpupload",
            # 'send-file-jingle',
            "start-video-call",
            "start-voice-call",
            "send-message",
            "show-contact-info",
            "show-chat-share",
            "muc-change-nickname",
            "muc-invite",
            "muc-contact-info",
            "muc-execute-command",
            "muc-ban",
            "muc-kick",
            "muc-change-role",
            "muc-change-affiliation",
            "muc-request-voice",
            "quote-next",
            "quote-prev",
        ]

        for action in actions:
            action = app.window.lookup_action(action)
            assert action is not None
            action.connect("activate", self._on_action)

        focus_action = app.window.lookup_action("focus-mode")
        assert focus_action is not None
        focus_action.bind_property(
            "state",
            self._banner_box,
            "visible",
            transform_to=transform_to_inverted_bool,
        )

    def _get_current_contact(self) -> ChatContactT:
        assert self._current_contact is not None
        return self._current_contact

    def process_escape(self) -> bool:
        if self.get_visible_child_name() == "function":
            self._chat_function_page.process_escape()
            return True

        if self._chat_control.process_escape():
            return True

        return self._message_action_box.process_escape()

    def get_chat_control(self) -> ChatControl:
        return self._chat_control

    def set_narrow(self, narrow: bool) -> None:
        self._chat_banner.set_narrow(narrow)
        self._chat_control.set_narrow(narrow)

    def get_message_action_box(self) -> MessageActionsBox:
        return self._message_action_box

    def get_message_input(self) -> MessageInputTextView:
        return self._message_action_box.get_message_input()

    def show_chat(self, account: str, jid: JID) -> None:
        # Store (preserve) primary clipboard and restore it after switching
        clipboard = self.get_primary_clipboard()
        if clipboard.get_content() is not None:
            clipboard.read_text_async(
                None, self._on_primary_clipboard_read_text_finished
            )

        self._last_quoted_id = None

        if self._current_contact is not None:
            self._current_contact.disconnect_all_from_obj(self)

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        self._current_contact = contact

        self._chat_banner.switch_contact(self._current_contact)
        self._chat_control.switch_contact(self._current_contact)
        self._message_action_box.switch_contact(self._current_contact)

        self._update_base_actions(self._current_contact)

        if isinstance(self._current_contact, GroupchatContact):
            self._current_contact.multi_connect(
                {
                    "user-joined": self._on_user_joined,
                    "user-role-changed": self._on_user_role_changed,
                    "user-affiliation-changed": self._on_user_affiliation_changed,
                    "state-changed": self._on_muc_state_changed,
                    "room-password-required": self._on_room_password_required,
                    "room-captcha-challenge": self._on_room_captcha_challenge,
                    "room-captcha-error": self._on_room_captcha_error,
                    "room-creation-failed": self._on_room_creation_failed,
                    "room-join-failed": self._on_room_join_failed,
                    "room-config-failed": self._on_room_config_failed,
                }
            )
            self._update_group_chat_actions(self._current_contact)

        elif isinstance(self._current_contact, GroupchatParticipant):
            self._update_participant_actions(self._current_contact)

        else:
            self._update_chat_actions(self._current_contact)

        if isinstance(self._current_contact, GroupchatContact):
            muc_data = client.get_module("MUC").get_muc_data(self._current_contact.jid)
            if muc_data is not None:
                if muc_data.state.is_captcha_request:
                    self._show_chat_function_page(FunctionMode.CAPTCHA_REQUEST)
                    return

                if muc_data.state.is_password_request:
                    self._show_chat_function_page(FunctionMode.PASSWORD_REQUEST)
                    return

                if not muc_data.state.is_joined:
                    if muc_data.error == "captcha-failed":
                        self._show_chat_function_page(
                            FunctionMode.CAPTCHA_ERROR, muc_data.error_text
                        )
                        return
                    if muc_data.error == "join-failed":
                        self._show_chat_function_page(
                            FunctionMode.JOIN_FAILED, muc_data.error_text
                        )
                        return
                    if muc_data.error == "creation-failed":
                        self._show_chat_function_page(
                            FunctionMode.CREATION_FAILED, muc_data.error_text
                        )
                        return

        self.set_transition_type(Gtk.StackTransitionType.NONE)
        self.set_visible_child_name("controls")

        app.plugin_manager.extension_point("switch_contact", self._current_contact)

        idle_add_once(app.window.activate_action, "win.input-focus")

    def _on_primary_clipboard_read_text_finished(
        self,
        clipboard: Gdk.Clipboard,
        result: Gio.AsyncResult,
    ) -> None:
        try:
            text = clipboard.read_text_finish(result)
        except Exception as e:
            InformationAlertDialog(_("Pasting Content Failed"), _("Error: %s") % e)
            return

        if text is None:
            log.info("No text pasted")
            return

        # Reset primary clipboard to what it was before switching chats,
        # otherwise it gets overridden.
        clipboard.set(text)

    def show_activity_page(self, context_id: str | None = None) -> None:
        if not context_id:
            self._activity_page.show_default_page()
        self.set_visible_child_name("activity")

    def _on_room_password_required(
        self, _contact: GroupchatContact, _signal_name: str
    ) -> None:
        self._show_chat_function_page(FunctionMode.PASSWORD_REQUEST)

    def _on_room_captcha_challenge(
        self, contact: GroupchatContact, _signal_name: str
    ) -> None:
        self._show_chat_function_page(FunctionMode.CAPTCHA_REQUEST)

    def _on_room_captcha_error(
        self, _contact: GroupchatContact, _signal_name: str, error: str
    ) -> None:
        self._show_chat_function_page(FunctionMode.CAPTCHA_ERROR, error)

    def _on_room_creation_failed(
        self, _contact: GroupchatContact, _signal_name: str, error: str
    ) -> None:
        self._show_chat_function_page(FunctionMode.CREATION_FAILED, error)

    def _on_room_join_failed(
        self, _contact: GroupchatContact, _signal_name: str, error: str
    ) -> None:
        self._show_chat_function_page(FunctionMode.JOIN_FAILED, error)

    def _on_room_config_failed(
        self, _contact: GroupchatContact, _signal_name: str, error: str
    ) -> None:
        self._show_chat_function_page(FunctionMode.CONFIG_FAILED)

    def _on_muc_state_changed(
        self, contact: GroupchatContact, _signal_name: str
    ) -> None:
        self._update_group_chat_actions(contact)

    def _on_user_joined(
        self,
        contact: GroupchatContact,
        _signal_name: str,
        _user_contact: GroupchatParticipant,
        _event: events.MUCUserJoined,
    ) -> None:
        self._update_group_chat_actions(contact)

    def _on_user_role_changed(
        self,
        contact: GroupchatContact,
        _signal_name: str,
        _user_contact: GroupchatParticipant,
        _event: events.MUCUserRoleChanged,
    ) -> None:
        self._update_group_chat_actions(contact)

    def _on_user_affiliation_changed(
        self,
        contact: GroupchatContact,
        _signal_name: str,
        _user_contact: GroupchatParticipant,
        _event: events.MUCUserAffiliationChanged,
    ) -> None:
        self._update_group_chat_actions(contact)

    def _on_muc_disco_update(self, event: events.MucDiscoUpdate) -> None:
        if not isinstance(self._current_contact, GroupchatContact):
            return

        if event.jid != self._current_contact.jid:
            return

        self._update_group_chat_actions(self._current_contact)

    def _on_feature_discovered(self, event: events.FeatureDiscovered) -> None:
        if self._current_contact is None:
            return

        if event.account != self._current_contact.account:
            return

        if event.feature == Namespace.HTTPUPLOAD_0:
            self._update_base_actions(self._current_contact)

    def _on_account_state(
        self, event: events.AccountConnected | events.AccountDisconnected
    ) -> None:
        if self._current_contact is None:
            return

        if event.account != self._current_contact.account:
            return

        self._update_base_actions(self._current_contact)

        if isinstance(self._current_contact, GroupchatContact):
            self._update_group_chat_actions(self._current_contact)
        elif isinstance(self._current_contact, GroupchatParticipant):
            self._update_participant_actions(self._current_contact)
        else:
            self._update_chat_actions(self._current_contact)

    def _on_message_received(self, event: events.MessageReceived) -> None:
        if event.from_mam:
            return

        if (
            app.window.is_chat_active(event.account, event.jid)
            and self._chat_control.view_is_at_bottom()
        ):
            if event.message.id is None:
                return

            client = app.get_client(event.account)
            stanza_id = event.message.stanza_id
            contact = client.get_module("Contacts").get_contact(event.jid)
            assert isinstance(
                contact, BareContact | GroupchatContact | GroupchatParticipant
            )
            mds_assist_sent = client.get_module("ChatMarkers").send_displayed_marker(
                contact, event.message.id, stanza_id
            )

            if not mds_assist_sent and stanza_id is not None:
                by = contact.jid if isinstance(contact, GroupchatContact) else None
                client.get_module("MDS").set_mds(contact.jid, stanza_id, by)

            return

        message = event.message
        if message.direction == ChatDirection.OUTGOING:
            return

        if event.m_type == MessageType.GROUPCHAT:
            client = app.get_client(event.account)
            contact = client.get_module("Contacts").get_contact(
                event.jid, groupchat=True
            )
            assert isinstance(contact, GroupchatContact)
            # MUC messages may be received after some delay, so make sure we
            # don't issue notifications for our own messages.
            self_contact = contact.get_self()
            if self_contact is not None and self_contact.name == message.resource:
                return

        self._issue_notification(event.account, message)

    def _issue_notification(self, account: str, message: Message) -> None:
        text = message.text
        assert text is not None

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(message.remote.jid)

        title = _("New message from")

        preview = get_preview_data(text, message.oob)
        if preview is not None:
            text = preview.text

        sound: str | None = None
        msg_type = "chat-message"
        if isinstance(contact, BareContact):
            msg_type = "chat-message"
            title += f" {contact.name}"
            sound = "first_message_received"
            app.window.set_urgency_hint(True)

        if isinstance(contact, GroupchatContact):
            msg_type = "group-chat-message"
            title += f" {message.resource} ({contact.name})"
            assert contact.nickname is not None
            needs_highlight = message_needs_highlight(
                text, contact.nickname, client.get_own_jid().bare
            )
            if needs_highlight:
                sound = "muc_message_highlight"
            else:
                sound = "muc_message_received"

            if not contact.can_notify() and not needs_highlight:
                return

            app.window.set_urgency_hint(True)

        if isinstance(contact, GroupchatParticipant):
            msg_type = "private-chat-message"
            title += f" {contact.name} (private in {contact.room.name})"
            sound = "first_message_received"

        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        if app.settings.get("notification_preview_message"):
            if text.startswith("/me "):
                name = contact.name
                if isinstance(contact, GroupchatContact):
                    name = message.resource
                text = f"* {name} {text[3:]}"
        else:
            text = ""

        if isinstance(contact, GroupchatContact):
            resource = message.resource
        else:
            resource = None

        app.ged.raise_event(
            events.Notification(
                context_id="",
                account=contact.account,
                jid=contact.jid,
                type="incoming-message",
                sub_type=msg_type,
                title=title,
                text=text,
                sound=sound,
                resource=resource,
            )
        )

    def _update_base_actions(self, contact: ChatContactT) -> None:
        client = app.get_client(contact.account)
        online = app.account_is_connected(contact.account)

        app.window.get_action("send-message").set_enabled(
            allow_send_message(self.get_message_input().has_text, contact)
        )

        httpupload = app.window.get_action("send-file-httpupload")
        httpupload.set_enabled(online and client.get_module("HTTPUpload").available)

        # jingle = app.window.get_action('send-file-jingle')
        # jingle.set_enabled(online and contact.is_jingle_available)

        app.window.get_action("send-file").set_enabled(
            # jingle.get_enabled() or
            httpupload.get_enabled()
        )

        app.window.get_action("correct-message").set_enabled(online)
        app.window.get_action("retract-message").set_enabled(online)

    def _update_chat_actions(self, contact: BareContact) -> None:
        online = app.account_is_connected(contact.account)

        app.window.get_action("start-voice-call").set_enabled(
            online and contact.supports_audio and sys.platform != "win32"
        )
        app.window.get_action("start-video-call").set_enabled(
            online and contact.supports_video and sys.platform != "win32"
        )

    def _update_group_chat_actions(self, contact: GroupchatContact) -> None:
        joined = contact.is_joined
        is_visitor = False
        if joined:
            self_contact = contact.get_self()
            assert self_contact
            is_visitor = self_contact.role.is_visitor

        app.window.get_action("muc-change-nickname").set_enabled(joined)
        app.window.get_action("muc-contact-info").set_enabled(joined)
        app.window.get_action("muc-execute-command").set_enabled(joined)
        app.window.get_action("muc-ban").set_enabled(joined)
        app.window.get_action("muc-kick").set_enabled(joined)
        app.window.get_action("muc-change-role").set_enabled(joined)
        app.window.get_action("muc-change-affiliation").set_enabled(joined)
        app.window.get_action("muc-invite").set_enabled(joined)

        app.window.get_action("muc-request-voice").set_enabled(is_visitor)

        app.window.get_action("moderate-message").set_enabled(joined)
        app.window.get_action("moderate-all-messages").set_enabled(joined)

    def _update_participant_actions(self, contact: GroupchatParticipant) -> None:
        pass

    def _on_action(self, action: Gio.SimpleAction, param: GLib.Variant | None) -> None:
        if self.get_visible_child_name() != "controls":
            return

        action_name = action.get_name()
        contact = self._current_contact
        if contact is None:
            return

        account = contact.account
        client = app.get_client(account)
        jid = contact.jid

        if action_name == "send-message":
            self._on_send_message()

        elif action_name == "start-voice-call":
            app.call_manager.start_call(account, jid, CallType.AUDIO)

        elif action_name == "start-video-call":
            app.call_manager.start_call(account, jid, CallType.VIDEO)

        elif action_name.startswith("send-file"):
            method = None
            name = action.get_name()
            if "httpupload" in name:
                method = "httpupload"

            # if 'jingle' in name:
            #     method = 'jingle'

            uris = None
            if param is not None:
                uris = param.get_strv()
                if all("" == uri for uri in param.get_strv()):
                    # All list entries are "" (default when invoking "send-file" action)
                    uris = None
            self._show_chat_function_page(FunctionMode.SEND_FILE, method, uris)

        elif action_name == "show-contact-info":
            if isinstance(contact, GroupchatContact):
                open_window("GroupchatDetails", contact=contact)
            else:
                open_window("ContactInfo", account=contact.account, contact=contact)

        elif action_name == "show-chat-share":
            self._chat_banner.present_share_popover()

        elif action_name == "muc-contact-info":
            assert param is not None
            nick = param.get_string()
            assert isinstance(contact, GroupchatContact)
            resource_contact = contact.get_resource(nick)
            open_window("ContactInfo", account=account, contact=resource_contact)

        elif action_name == "muc-invite":
            self._show_chat_function_page(FunctionMode.INVITE)

        elif action_name == "muc-change-nickname":
            self._show_chat_function_page(FunctionMode.CHANGE_NICKNAME)

        elif action_name == "muc-execute-command":
            nick = None
            if param is not None:
                nick = param.get_string()
            if nick:
                assert isinstance(contact, GroupchatContact)
                resource_contact = contact.get_resource(nick)
                jid = resource_contact.jid
            open_window("AdHocCommands", account=account, jids=[jid])

        elif action_name == "muc-kick":
            assert param is not None
            kick_nick = param.get_string()
            self._show_chat_function_page(FunctionMode.KICK, data=kick_nick)

        elif action_name == "muc-ban":
            assert param is not None
            ban_jid = param.get_string()
            self._show_chat_function_page(FunctionMode.BAN, data=ban_jid)

        elif action_name == "muc-change-role":
            assert param is not None
            assert isinstance(contact, GroupchatContact)
            nick, role = param.get_strv()
            client.get_module("MUC").set_role(
                contact.jid,
                nick,
                role,
                callback=self._on_affiliation_or_role_change,
                user_data=(contact, jid, role),
            )

        elif action_name == "muc-change-affiliation":
            assert param is not None
            assert isinstance(contact, GroupchatContact)
            jid, affiliation = param.get_strv()
            client.get_module("MUC").set_affiliation(
                contact.jid,
                {jid: {"affiliation": affiliation}},
                callback=self._on_affiliation_or_role_change,
                user_data=(contact, jid, affiliation),
            )

        elif action_name == "muc-request-voice":
            client.get_module("MUC").request_voice(contact.jid)

        elif action_name.startswith("quote-"):
            view = self._chat_control.get_conversation_view()
            if action_name == "quote-prev":
                row = view.get_prev_message_row(self._last_quoted_id)
            else:
                row = view.get_next_message_row(self._last_quoted_id)

            if row is not None:
                self._last_quoted_id = row.pk
                self._message_action_box.insert_as_quote(row.get_text(), clear=True)

    def _on_affiliation_or_role_change(self, task: Task) -> None:
        muc, jid, affiliation_or_role = task.get_user_data()

        try:
            result = task.finish()
        except StanzaError as error:
            log.error("Error on affiliation/role change request: %s", error)
            if self._current_contact == muc:
                self._chat_control.add_info_message(
                    _(
                        "An error occurred while trying to make "
                        "{occupant_jid} {affiliation_or_role}: {error}"
                    ).format(
                        occupant_jid=jid,
                        affiliation_or_role=affiliation_or_role,
                        error=str(error),
                    )
                )
            else:
                error_message = (
                    _(
                        "An error occurred while trying to make "
                        "{occupant_jid} {affiliation_or_role} in group "
                        '"{group}"'
                    ).format(
                        occupant_jid=jid,
                        affiliation_or_role=affiliation_or_role,
                        group=muc.name,
                    ),
                )
                InformationAlertDialog(_("Error"), f"{error_message}\n{error}")
        else:
            log.debug("Affiliation/role change success: %s", result)

    def _on_drag_enter(
        self,
        _drop_target: Gtk.DropTarget,
        _x: float,
        _y: float,
    ) -> Gdk.DragAction:
        self._drop_area.set_visible(True)
        return Gdk.DragAction.COPY

    def _on_drag_leave(self, _drop_target: Gtk.DropTarget) -> None:
        self._drop_area.set_visible(False)

    def _on_drop_accept(self, target: Gtk.DropTarget, drop: Gdk.Drop) -> bool:
        # Accept handler needs to be here, the default accept handler returns
        # always False, because the overlay is not visible
        allowed_formats = target.get_formats()
        assert allowed_formats is not None

        offered_formats = drop.get_formats()

        log.debug("Offered formats for drop: %s", offered_formats.get_gtypes())
        if not allowed_formats.match(offered_formats):
            return False

        # DND on X11 freezes due to a GTK bug:
        # https://dev.gajim.org/gajim/gajim/-/issues/12313
        if app.is_display(Display.X11):
            return gi_gui_package_version("Gtk>=4.20.1")

        return True

    def _on_file_drop(
        self,
        target: Gtk.DropTargetAsync,
        value: Gdk.FileList | Gdk.Texture | None,
        _x: float,
        _y: float,
    ) -> bool:
        if value is None:
            log.debug("Drop received, but value is None")
            return False

        if not self._chat_control.has_active_chat():
            return False

        files: list[str] = []

        log.debug("Drop received: %s", value)
        if isinstance(value, Gdk.Texture):
            try:
                image = value.save_to_png_bytes()
                image_bytes = image.get_data()
            except GLib.Error as e:
                log.error("Error while trying to store dropped image: %s", e)
                return False

            if image_bytes is None:
                log.error("Could not process dropped image")
                return False

            temp_dir = configpaths.get_temp_dir()
            image_path = temp_dir / f"{uuid.uuid4()}.png"
            try:
                image_path.write_bytes(image_bytes)
            except Exception as e:
                log.error("Could not write dropped image: %s", e)
                return False

            files = [image_path.as_uri()]

        if isinstance(value, Gdk.FileList):
            files = [file.get_uri() for file in value.get_files()]

        if not files:
            log.error("No path for dropped image")
            return False

        self._chat_control.drag_data_file_transfer(files)
        self._drop_area.set_visible(False)
        return True

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if not self._chat_control.message_selection_active():
            return Gdk.EVENT_PROPAGATE

        if state & Gdk.ModifierType.CONTROL_MASK:
            if keyval == Gdk.KEY_c:
                self._chat_control.copy_selected_messages()
                return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _show_chat_function_page(
        self,
        function_mode: FunctionMode,
        data: str | None = None,
        files: list[str] | None = None,
    ) -> None:
        assert self._current_contact is not None
        self._chat_function_page.set_mode(
            self._current_contact, function_mode, data, files
        )
        self.set_transition_type(Gtk.StackTransitionType.SLIDE_UP_DOWN)
        self.set_visible_child_name("function")

    def _on_function_finished(
        self, function_page: ChatFunctionPage, close_control: bool
    ) -> None:
        if close_control:
            self._close_control()
            self.set_visible_child_name("empty")
            self.set_transition_type(Gtk.StackTransitionType.NONE)
            function_page.reset()
            return

        self.set_visible_child_name("controls")
        self.set_transition_type(Gtk.StackTransitionType.NONE)
        function_page.reset()

    def _on_function_message(
        self, _function_page: ChatFunctionPage, message: str
    ) -> None:
        self._chat_control.add_info_message(message)

    def check_send_preconditons(self) -> bool:
        contact = self._current_contact
        assert contact is not None

        client = app.get_client(contact.account)

        encryption = contact.settings.get("encryption")
        if not encryption:
            return True

        if encryption == "OMEMO":
            return client.get_module("OMEMO").check_send_preconditions(contact)

        if encryption == "OpenPGP":
            if not client.get_module("OpenPGP").secret_key_exists():
                open_window("OpenPGPWizard", account=contact.account)
                return False

            return client.get_module("OpenPGP").check_send_preconditions(contact)

        if encryption not in app.plugin_manager.encryption_plugins:
            InformationAlertDialog(
                _("Encryption Error"), _("Missing necessary encryption plugin")
            )
            return False

        self._chat_control.sendmessage = True
        app.plugin_manager.extension_point(
            "send_message" + encryption, self._chat_control
        )
        return self._chat_control.sendmessage

    def _on_send_message(self) -> None:
        message = self._message_action_box.get_text()
        if message.startswith("//"):
            # Escape sequence for chat commands
            message = message[1:]

        contact = self._current_contact
        assert contact is not None

        client = app.get_client(contact.account)

        if not self.check_send_preconditons():
            return

        message = remove_invalid_xml_chars(message)
        if message in ("", "\n"):
            return

        label = self._message_action_box.get_seclabel()
        correct_id = self._message_action_box.get_correction_id()
        reply_data = self._message_action_box.get_message_reply()
        open_graph_data = self._message_action_box.get_open_graph_data()

        chatstate = client.get_module("Chatstate").get_active_chatstate(contact)

        message_ = OutgoingMessage(
            account=contact.account,
            contact=contact,
            text=message,
            chatstate=chatstate,
            sec_label=label,
            control=self._chat_control,
            correct_id=correct_id,
            reply_data=reply_data,
            open_graph_data=open_graph_data,
        )

        client.send_message(message_)

        self._message_action_box.reset_state_after_send()

        self._last_quoted_id = None

    def get_last_message_id(self, contact: ChatContactT) -> str | None:
        return self._message_action_box.get_last_message_id(contact)

    def _close_control(self) -> None:
        assert self._current_contact is not None
        app.window.activate_action(
            "win.remove-chat",
            GLib.Variant(
                "as", [self._current_contact.account, str(self._current_contact.jid)]
            ),
        )

    def clear(self) -> None:
        if self._current_contact is not None:
            self._current_contact.disconnect_all_from_obj(self)

        self._last_quoted_id = None
        self.set_visible_child_name("empty")
        self._chat_banner.clear()
        self._message_action_box.clear()
        self._chat_control.clear()

    def _on_command_signal(
        self, _chat_commands: ChatCommands, signal_name: str, text: str
    ) -> None:
        is_error = signal_name != "command-result"
        self._chat_control.add_command_output(text, is_error)


class ChatPlaceholderBox(Gtk.Box):
    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=18)
        self.set_valign(Gtk.Align.CENTER)
        image = Gtk.Image.new_from_icon_name("gajim-symbolic")
        image.set_pixel_size(100)
        image.set_opacity(0.2)
        self.append(image)

        button = Gtk.Button(
            label=_("Start Chatting…"),
            halign=Gtk.Align.CENTER,
            action_name="app.start-chat",
            action_target=GLib.Variant("as", ["", ""]),
        )
        self.append(button)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Concatenate
from typing import ParamSpec
from typing import TypeVar

import dataclasses
import json
import logging
import math
import os
import pprint
import sqlite3
import sys
import time
from collections.abc import Callable
from collections.abc import Iterator
from datetime import datetime
from datetime import UTC
from pathlib import Path

import nbxmpp.const
import sqlalchemy as sa
import sqlalchemy.exc
from gi.repository import GLib
from nbxmpp.const import Affiliation
from nbxmpp.const import Role
from nbxmpp.const import StatusCode
from nbxmpp.modules.discovery import parse_disco_info
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.structs import CommonError
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import RosterItem
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import DBAPIConnection
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import ORMExecuteState
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.expression import ClauseElement
from sqlalchemy.sql.expression import Executable

from gajim.common.const import Draft
from gajim.common.const import ValueMissingT
from gajim.common.util.version import python_version

log = logging.getLogger("gajim.c.storage")

P = ParamSpec("P")
R = TypeVar("R")


def timeit(func: Callable[P, R]) -> Callable[P, R]:
    def func_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        if log.getEffectiveLevel() != logging.DEBUG:
            return func(*args, **kwargs)

        start = time.time()
        result = func(*args, **kwargs)
        exec_time = (time.time() - start) * 1e3
        log.debug("Execution time for %s: %s ms", func.__name__, math.ceil(exec_time))
        return result

    return func_wrapper


def with_session(
    func: Callable[Concatenate[Any, Session, P], R],
) -> Callable[Concatenate[Any, P], R]:
    def wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> R:
        with self._create_session() as session, session.begin():
            return func(self, session, *args, **kwargs)

    return wrapper


def with_session_yield_from(
    func: Callable[Concatenate[Any, Session, P], Iterator[R]],
) -> Callable[Concatenate[Any, P], Iterator[Any]]:
    def wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> Iterator[R]:
        with self._create_session() as session, session.begin():
            yield from func(self, session, *args, **kwargs)

    return wrapper


def _convert_common_error(common_error: bytes) -> CommonError:
    return CommonError.from_string(common_error)


def _adapt_common_error(common_error: CommonError) -> str:
    return common_error.serialize()


sqlite3.register_converter("common_error", _convert_common_error)
sqlite3.register_adapter(CommonError, _adapt_common_error)


def _convert_marker(marker: bytes):
    return "received" if int(marker) == 0 else "displayed"


sqlite3.register_converter("marker", _convert_marker)


def _jid_adapter(jid: JID) -> str:
    return str(jid)


def _jid_converter(jid: bytes) -> JID:
    return JID.from_string(jid.decode())


sqlite3.register_converter("jid", _jid_converter)
sqlite3.register_adapter(JID, _jid_adapter)


def _convert_disco_info(disco_info: bytes) -> DiscoInfo:
    return parse_disco_info(Iq(node=disco_info))  # type: ignore


def _adapt_disco_info(disco_info: DiscoInfo) -> str:
    return str(disco_info.stanza)


sqlite3.register_converter("disco_info", _convert_disco_info)
sqlite3.register_adapter(DiscoInfo, _adapt_disco_info)


def _convert_json(json_string: bytes) -> dict[str, Any]:
    return json.loads(json_string, object_hook=json_decoder)


sqlite3.register_converter("JSON", _convert_json)


def _datetime_converter(data: bytes) -> datetime:
    return datetime.fromisoformat(data.decode())


sqlite3.register_converter("datetime", _datetime_converter)


sqlite3.register_adapter(ValueMissingT, lambda _val: None)


class Encoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, set):
            return list(o)  # type: ignore

        if isinstance(o, datetime):
            return {"__type": "datetime", "value": o.isoformat()}

        if isinstance(o, JID):
            return {"__type": "JID", "value": str(o)}

        if isinstance(o, RosterItem):
            dct = o.asdict()
            dct["__type"] = "RosterItem"
            return dct

        if isinstance(o, Affiliation | Role | StatusCode):
            return {"value": o.value, "__type": o.__class__.__name__}

        return json.JSONEncoder.default(self, o)


def json_decoder(dct: dict[str, Any]) -> Any:
    type_ = dct.get("__type")
    if type_ is None:
        return dct

    if type_ == "JID":
        return JID.from_string(dct["value"])

    if type_ == "datetime":
        return datetime.fromisoformat(dct["value"])

    if type_ == "RosterItem":
        return RosterItem(
            jid=dct["jid"],
            name=dct["name"],
            ask=dct["ask"],
            subscription=dct["subscription"],
            approved=dct["approved"],
            groups=set(dct["groups"]),
        )

    if type_ in ("Affiliation", "Role", "StatusCode"):
        return getattr(nbxmpp.const, type_)(dct["value"])

    return dct


class Explain(Executable, ClauseElement):
    inherit_cache = False

    def __init__(self, stmt: Executable) -> None:
        self.statement = stmt


@compiles(Explain, "sqlite")
def sqlite_explain(element: Any, compiler: SQLCompiler, **kw: Any):
    compiled = compiler.process(element.statement, **kw)
    return f"EXPLAIN QUERY PLAN {compiled}"


def _do_orm_execute(orm_execute_state: ORMExecuteState) -> None:
    stmt = orm_execute_state.statement
    if not stmt.is_select and not stmt.is_delete:
        return

    res = orm_execute_state.session.execute(
        Explain(stmt), params=orm_execute_state.parameters
    ).all()
    explanation = pprint.pformat(res)
    log.debug("EXPLANATION\n%s\n%s", explanation, stmt)


if os.environ.get("GAJIM_DEBUG_SQL"):
    event.listen(Session, "do_orm_execute", _do_orm_execute)


class SqliteStorage:
    """
    Base Storage Class
    """

    def __init__(
        self,
        log: logging.Logger,
        path: Path | None,
        create_statement: str,
        commit_delay: int = 500,
    ) -> None:
        self._log = log
        self._path = path
        self._create_statement = create_statement
        self._commit_delay = commit_delay
        self._con = cast(sqlite3.Connection, None)
        self._commit_source_id = None

    def init(self, **kwargs: Any) -> None:
        if self._path is None or not self._path.exists():
            self._con = self._create_storage(**kwargs)

        else:
            if not self._path.is_file():
                sys.exit("%s must be a file" % self._path)
            self._con = self._connect(**kwargs)

        self._migrate_storage()

    def get_connection(self) -> sqlite3.Connection:
        # Use this only for unittests
        return self._con

    def _enable_foreign_keys(self) -> None:
        self._con.execute("PRAGMA foreign_keys=ON")

    def _set_journal_mode(self, mode: str) -> None:
        self._con.execute(f"PRAGMA journal_mode={mode}")

    def _set_synchronous(self, mode: str) -> None:
        self._con.execute(f"PRAGMA synchronous={mode}")

    def _enable_secure_delete(self) -> None:
        self._con.execute("PRAGMA secure_delete=1")

    def _run_optimize(self) -> None:
        self._con.execute("PRAGMA optimize")

    @property
    def user_version(self) -> int:
        return self._con.execute("PRAGMA user_version").fetchone()[0]

    def _connect(self, **kwargs: Any) -> sqlite3.Connection:
        self._log.info("Connect to %s", self._path)
        return sqlite3.connect(self._path or ":memory:", **kwargs)

    def _create_storage(self, **kwargs: Any) -> sqlite3.Connection:
        self._log.info("Creating %s", self._path or "in memory")

        con = self._connect(**kwargs)

        if self._path is not None:
            self._path.chmod(0o600)

        try:
            con.executescript(self._create_statement)
        except Exception:
            self._log.exception("Error")
            con.close()
            if self._path is not None:
                self._path.unlink()
            sys.exit("Failed creating storage")

        con.commit()
        return con

    def _reinit_storage(self) -> None:
        self._con.close()
        if self._path is not None:
            self._path.unlink()
        self.init()

    def _migrate_storage(self) -> None:
        try:
            self._migrate()
        except Exception:
            self._con.close()
            self._log.exception("Error")
            sys.exit()

    def _migrate(self) -> None:
        raise NotImplementedError

    def _execute_multiple(self, statements: list[str]) -> None:
        """
        Execute multiple statements with the option to fail on duplicates
        but still continue
        """
        for sql in statements:
            try:
                self._con.execute(sql)
                self._con.commit()
            except sqlite3.OperationalError as error:
                if str(error).startswith("duplicate column name:"):
                    self._log.info(error)
                else:
                    self._con.close()
                    self._log.exception("Error")
                    sys.exit()

    @timeit
    def _commit(self) -> bool:
        self._commit_source_id = None
        self._con.commit()
        return False

    def _delayed_commit(self) -> None:
        if self._commit_source_id is not None:
            return

        self._commit_source_id = GLib.timeout_add(self._commit_delay, self._commit)

    def shutdown(self) -> None:
        if self._commit_source_id is not None:
            GLib.source_remove(self._commit_source_id)

        self._commit()
        self._run_optimize()
        self._con.close()
        del self._con


class AlchemyStorage:
    def __init__(
        self,
        log: logging.Logger,
        path: Path | None,
        pragma: dict[str, str] | None = None,
    ) -> None:
        self._log = log
        self._path = path
        self._engine = self._create_engine()
        self._commit_source_id = None
        self._pragma = pragma or {}

    def init(self) -> None:
        if self._path is None or not self._path.exists():
            self._create_storage()

        elif not self._path.is_file():
            sys.exit("%s must be a file" % self._path)

        self._migrate_storage()

    def get_session(self) -> Session:
        return self._create_session()

    def get_engine(self) -> Engine:
        return self._engine

    def _dbapi_execute_multiple(
        self,
        statements: list[str],
        *,
        dbapi_connection: DBAPIConnection | None = None,
    ) -> None:
        if dbapi_connection is None:
            dbapi_connection = self._engine.raw_connection().dbapi_connection
            assert dbapi_connection is not None

        # the sqlite3 driver will not set PRAGMA if autocommit=False
        # also VACUUM does not work with
        dbapi_connection.autocommit = True

        cursor = dbapi_connection.cursor()
        for stmt in statements:
            cursor.execute(stmt)
        cursor.close()

        dbapi_connection.autocommit = False

    def set_user_version(self, version: int) -> None:
        self._dbapi_execute_multiple([f"PRAGMA user_version={version}"])

    def set_foreign_keys(self, enabled: bool) -> None:
        self._dbapi_execute_multiple([f"PRAGMA foreign_keys={int(enabled)}"])

    def _on_db_connect(
        self, dbapi_connection: DBAPIConnection, _connection_record: Any
    ) -> None:
        stmts = ["PRAGMA foreign_keys=1"]
        stmts += [f"PRAGMA {key}={value}" for key, value in self._pragma.items()]
        self._dbapi_execute_multiple(stmts, dbapi_connection=dbapi_connection)

    def run_analyze(self) -> None:
        self._dbapi_execute_multiple(["PRAGMA analysis_limit=0", "ANALYZE"])

    def _run_optimize(self) -> None:
        self._dbapi_execute_multiple(["PRAGMA optimize", "VACUUM"])

    def get_user_version(self) -> int:
        with self._create_session() as s:
            return s.scalar(sa.text("PRAGMA user_version"))

    def _create_engine(self) -> Engine:
        self._log.info("Create engine")

        con_str = "sqlite://"
        if self._path is not None:
            con_str += f"/{self._path}"

        engine = sa.create_engine(
            con_str, connect_args={"check_same_thread": False}, echo=False
        )
        event.listen(engine, "connect", self._on_db_connect)
        return engine

    def _create_session(self) -> Session:
        return sessionmaker(
            expire_on_commit=False, autoflush=False, bind=self._engine
        )()

    def _create_storage(self) -> None:
        self._log.info("Creating %s", self._path or "in memory")

        with self._create_session() as s:
            try:
                self._create_table(s, self._engine)
            except Exception:
                self._log.exception("Error")
                if self._path is not None:
                    self._path.unlink()
                sys.exit("Failed creating storage")

        if self._path is not None:
            self._path.chmod(0o600)

    def _create_table(self, session: Session, engine: Engine) -> None:
        raise NotImplementedError

    def _reinit_storage(self) -> None:
        self._engine.dispose()
        if self._path is not None:
            self._path.unlink()
        self._engine = self._create_engine()
        self.init()

    def _migrate_storage(self) -> None:
        try:
            self._migrate()
        except Exception:
            self._log.exception("Migration error")
            raise

    def _migrate(self) -> None:
        raise NotImplementedError

    @with_session
    @timeit
    def refresh(self, session: Session, obj: Any, attrs: list[str]) -> None:
        session.add(obj)
        session.refresh(obj, attribute_names=attrs)

    def shutdown(self) -> None:
        self._run_optimize()
        self._engine.dispose()
        del self._engine


class JIDType(sa.types.TypeDecorator[JID]):
    impl = sa.types.TEXT
    cache_ok = True

    def process_bind_param(self, value: JID | None, dialect: Any) -> str | None:
        if value is None:
            return value
        return str(value)

    def process_result_value(self, value: str | None, dialect: Any) -> JID | None:
        if value is None:
            return value
        return JID.from_string(value)


class StrValueMissingType(sa.types.TypeDecorator[Any]):
    impl = sa.types.TEXT
    cache_ok = True

    def process_bind_param(
        self, value: str | ValueMissingT | None, dialect: Any
    ) -> str | None:
        if isinstance(value, ValueMissingT):
            return None
        return value


class EpochTimestampType(sa.types.TypeDecorator[Any]):
    impl = sa.types.FLOAT
    cache_ok = True

    def process_bind_param(
        self, value: datetime | ValueMissingT | None, dialect: Any
    ) -> float | None:
        if value is None or isinstance(value, ValueMissingT):
            return None

        if value.tzinfo != UTC:
            raise ValueError("DateTime must be UTC")
        return value.timestamp()

    def process_result_value(
        self, value: float | None, dialect: Any
    ) -> datetime | None:
        if value is None:
            return None
        return datetime.fromtimestamp(value, UTC)


class JSONType(sa.types.TypeDecorator[Any]):
    impl = sa.types.TEXT
    cache_ok = True

    def process_bind_param(self, value: dict[str, Any] | None, dialect: Any):
        if value is not None:
            return json.dumps(value)
        return value

    def process_result_value(self, value: str | None, dialect: Any):
        if value is not None:
            return json.loads(value)
        return value


class DraftType(sa.types.TypeDecorator[Any]):
    impl = sa.types.TEXT
    cache_ok = True

    def process_bind_param(
        self, value: Draft | ValueMissingT | None, dialect: Any
    ) -> str | None:
        if isinstance(value, ValueMissingT) or value is None:
            return None
        return json.dumps(dataclasses.asdict(value))

    def process_result_value(self, value: str | None, dialect: Any) -> Draft | None:
        if value is None:
            return None
        return Draft(**json.loads(value))


def is_unique_constraint_error(error: sqlalchemy.exc.DatabaseError) -> bool:
    if not isinstance(error, sqlalchemy.exc.IntegrityError):
        return False

    if python_version("<3.11"):
        return "UNIQUE constraint failed" in error.args[0]
    return (
        error.orig.sqlite_errorcode  # type: ignore
        == sqlite3.SQLITE_CONSTRAINT_UNIQUE
    )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk


class QuoteWidget(Gtk.Box):
    def __init__(self, account: str) -> None:
        Gtk.Box.__init__(self)
        self.set_vexpand(True)
        self.add_css_class("conversation-quote")
        quote_bar = Gtk.Box()
        quote_bar.set_size_request(3, -1)
        quote_bar.set_margin_end(6)
        quote_bar.add_css_class("conversation-quote-bar")
        self.append(quote_bar)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import NamedTuple

from enum import IntEnum

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.const import StyleAttr
from gajim.common.events import StyleChanged
from gajim.common.events import ThemeUpdate
from gajim.common.i18n import _

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import iterate_listbox_children
from gajim.gtk.util.window import get_app_window
from gajim.gtk.window import GajimAppWindow


class StyleOption(NamedTuple):
    label: str
    selector: str
    attr: StyleAttr


CSS_STYLE_OPTIONS: list[StyleOption] = [
    StyleOption(
        _("Conversation: Text Font"), ".gajim-conversation-text", StyleAttr.FONT
    ),
    StyleOption(
        _("Conversation: Text Color"), ".gajim-conversation-text", StyleAttr.COLOR
    ),
    StyleOption(
        _("Conversation: Nickname Color (Incoming)"),
        ".gajim-incoming-nickname",
        StyleAttr.COLOR,
    ),
    StyleOption(
        _("Conversation: Nickname Color (Outgoing)"),
        ".gajim-outgoing-nickname",
        StyleAttr.COLOR,
    ),
    StyleOption(
        _("Status Message: Text Color"), ".gajim-status-message", StyleAttr.COLOR
    ),
    StyleOption(
        _("Status Message: Text Font"), ".gajim-status-message", StyleAttr.FONT
    ),
    StyleOption(_("Chat Banner: Foreground Color"), ".gajim-banner", StyleAttr.COLOR),
    StyleOption(
        _("Chat Banner: Background Color"), ".gajim-banner", StyleAttr.BACKGROUND
    ),
    StyleOption(_("Chat Banner: Text Font"), ".gajim-banner", StyleAttr.FONT),
    StyleOption(_("Status: Online Color"), ".gajim-status-online", StyleAttr.COLOR),
    StyleOption(_("Status: Away Color"), ".gajim-status-away", StyleAttr.COLOR),
    StyleOption(_("Status: DND Color"), ".gajim-status-dnd", StyleAttr.COLOR),
    StyleOption(_("Status: Offline Color"), ".gajim-status-offline", StyleAttr.COLOR),
]


class Column(IntEnum):
    THEME = 0


class Themes(GajimAppWindow):
    def __init__(self, transient: Gtk.Window) -> None:
        GajimAppWindow.__init__(
            self,
            name="Themes",
            title=_("Gajim Themes"),
            default_width=600,
            default_height=400,
            transient_for=transient,
            modal=True,
            add_window_padding=True,
            header_bar=True,
        )

        self._ui = get_builder("themes_window.ui")
        self.set_child(self._ui.theme_grid)

        self._get_themes()
        self._ui.option_listbox.set_placeholder(self._ui.placeholder)

        self._fill_choose_listbox()

        self._connect(self._ui.choose_option_listbox, "row-activated", self._add_option)
        self._connect(
            self._ui.theme_treeview_selection, "changed", self._on_theme_selected
        )
        self._connect(
            self._ui.theme_name_cell_renderer, "edited", self._on_theme_name_edit
        )
        self._connect(self._ui.add_theme_button, "clicked", self._on_add_new_theme)
        self._connect(self._ui.remove_theme_button, "clicked", self._on_remove_theme)

    def _get_themes(self) -> None:
        current_theme = app.settings.get("roster_theme")
        for theme in app.css_config.themes:
            if theme == current_theme:
                self._ui.theme_store.prepend([theme])
                continue
            self._ui.theme_store.append([theme])

    def _on_theme_name_edit(
        self, _renderer: Gtk.CellRendererText, path: str, new_name: str
    ) -> None:
        iter_ = self._ui.theme_store.get_iter(path)
        old_name = self._ui.theme_store[iter_][Column.THEME]

        if new_name == "default":
            InformationAlertDialog(
                _("Invalid Name"),
                _("Name <b>default</b> is not allowed"),
            )
            return

        if " " in new_name:
            InformationAlertDialog(_("Invalid Name"), _("Spaces are not allowed"))
            return

        if new_name == "":
            return

        result = app.css_config.rename_theme(old_name, new_name)
        if result is False:
            return

        app.settings.set("roster_theme", new_name)
        self._ui.theme_store.set_value(iter_, Column.THEME, new_name)

    def _select_theme_row(self, iter_: Gtk.TreeIter) -> None:
        self._ui.theme_treeview.get_selection().select_iter(iter_)

    def _on_theme_selected(self, tree_selection: Gtk.TreeSelection) -> None:
        store, iter_ = tree_selection.get_selected()
        if iter_ is None:
            self._clear_options()
            return
        theme = store[iter_][Column.THEME]
        app.css_config.change_preload_theme(theme)

        self._ui.remove_theme_button.set_sensitive(True)
        self._load_options()
        self._apply_theme(theme)
        app.ged.raise_event(StyleChanged())

    def _load_options(self) -> None:
        self._clear_options()

        for option in CSS_STYLE_OPTIONS:
            value = app.css_config.get_value(option.selector, option.attr, pre=True)

            if value is None:
                continue

            row = Option(option, value)
            self._ui.option_listbox.append(row)

    def _add_option(self, _listbox: Gtk.ListBox, row: Option) -> None:
        # Add theme if there is none
        store, _ = self._ui.theme_treeview.get_selection().get_selected()
        first = store.get_iter_first()
        if first is None:
            self._on_add_new_theme()

        # Don't add an option twice
        for option in iterate_listbox_children(self._ui.option_listbox):
            if option == row:
                return

        # Get default value if it exists
        value = app.css_config.get_value(row.option.selector, row.option.attr)

        row = Option(row.option, value)
        self._ui.option_listbox.append(row)
        self._ui.option_popover.popdown()

    def _clear_options(self) -> None:
        for row in iterate_listbox_children(self._ui.option_listbox):
            self._ui.option_listbox.remove(row)

    def _fill_choose_listbox(self) -> None:
        for option in CSS_STYLE_OPTIONS:
            self._ui.choose_option_listbox.append(ChooseOption(option))

    def _on_add_new_theme(self, *args: Any) -> None:
        name = self._create_theme_name()
        if not app.css_config.add_new_theme(name):
            return

        self._ui.remove_theme_button.set_sensitive(True)
        iter_ = self._ui.theme_store.append([name])
        self._select_theme_row(iter_)
        self._apply_theme(name)

    def reload_roster_theme(self) -> None:
        tree_selection = self._ui.theme_treeview.get_selection()
        store, iter_ = tree_selection.get_selected()
        if iter_ is None:
            return
        theme = store[iter_][Column.THEME]
        self._apply_theme(theme)

    @staticmethod
    def _apply_theme(theme: str) -> None:
        app.settings.set("roster_theme", theme)
        app.css_config.change_theme(theme)
        app.ged.raise_event(ThemeUpdate())

    @staticmethod
    def _create_theme_name() -> str:
        index = 0
        while f"newtheme{index}" in app.css_config.themes:
            index += 1
        return f"newtheme{index}"

    def _on_remove_theme(self, *args: Any) -> None:
        store, iter_ = self._ui.theme_treeview.get_selection().get_selected()
        if iter_ is None:
            return

        theme = store[iter_][Column.THEME]

        def _on_response() -> None:
            if theme == app.settings.get("roster_theme"):
                self._apply_theme("default")

            app.css_config.remove_theme(theme)
            app.ged.raise_event(ThemeUpdate())

            assert isinstance(store, Gtk.ListStore)
            assert iter_ is not None
            store.remove(iter_)

            first = store.get_iter_first()
            if first is None:
                self._ui.remove_theme_button.set_sensitive(False)
                self._clear_options()

        text = _("Do you want to delete this theme?")
        if theme == app.settings.get("roster_theme"):
            text = _(
                "This is the theme you are currently using.\n"
                "Do you want to delete this theme?"
            )

        ConfirmationAlertDialog(
            _("Delete Theme?"),
            text,
            confirm_label=_("_Delete"),
            appearance="destructive",
            callback=_on_response,
            parent=self,
        )

    def _cleanup(self) -> None:
        pass


class Option(Gtk.ListBoxRow, SignalManager):  # noqa: PLW1641
    def __init__(
        self, option: StyleOption, value: str | Pango.FontDescription | None
    ) -> None:
        Gtk.ListBoxRow.__init__(self)
        SignalManager.__init__(self)

        self.option = option
        self._box = Gtk.Box(spacing=12)

        label = Gtk.Label()
        label.set_text(option.label)
        label.set_hexpand(True)
        label.set_halign(Gtk.Align.START)
        self._box.append(label)

        if option.attr in (StyleAttr.COLOR, StyleAttr.BACKGROUND):
            assert not isinstance(value, Pango.FontDescription)
            self._init_color(value)
        elif option.attr == StyleAttr.FONT:
            assert not isinstance(value, str)
            self._init_font(value)

        remove_button = Gtk.Button.new_from_icon_name("lucide-trash-symbolic")
        remove_button.set_tooltip_text(_("Remove Setting"))
        self._connect(remove_button, "clicked", self._on_remove)
        self._box.append(remove_button)

        self.set_child(self._box)

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def _init_color(self, color: str | None) -> None:
        color_dialog = Gtk.ColorDialog()
        color_button = Gtk.ColorDialogButton(dialog=color_dialog)
        if color is not None:
            rgba = Gdk.RGBA()
            rgba.parse(color)
            color_button.set_rgba(rgba)
        color_button.set_halign(Gtk.Align.END)
        self._connect(color_button, "notify::rgba", self._on_color_set)
        self._box.append(color_button)

    def _init_font(self, desc: Pango.FontDescription | None) -> None:
        font_dialog = Gtk.FontDialog()
        font_button = Gtk.FontDialogButton(dialog=font_dialog)
        if desc is not None:
            font_button.set_font_desc(desc)
        font_button.set_halign(Gtk.Align.END)
        self._connect(font_button, "notify::font-desc", self._on_font_set)
        self._box.append(font_button)

    def _on_color_set(self, color_button: Gtk.ColorDialogButton, *args: Any) -> None:
        color = color_button.get_rgba()
        color_string = color.to_string()
        app.css_config.set_value(
            self.option.selector, self.option.attr, color_string, pre=True
        )
        app.ged.raise_event(StyleChanged())

        themes_win = get_app_window("Themes")
        assert themes_win is not None
        themes_win.reload_roster_theme()

    def _on_font_set(self, font_button: Gtk.FontDialogButton, *args: Any) -> None:
        desc = font_button.get_font_desc()
        if desc is None:
            return
        app.css_config.set_font(self.option.selector, desc, pre=True)
        app.ged.raise_event(StyleChanged())

        themes_win = get_app_window("Themes")
        assert themes_win is not None
        themes_win.reload_roster_theme()

    def _on_remove(self, *args: Any) -> None:
        listbox = cast(Gtk.ListBox, self.get_parent())
        listbox.remove(self)
        app.css_config.remove_value(self.option.selector, self.option.attr, pre=True)
        app.ged.raise_event(StyleChanged())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ChooseOption):
            raise NotImplementedError
        return other.option == self.option


class ChooseOption(Gtk.ListBoxRow):
    def __init__(self, option: StyleOption) -> None:
        Gtk.ListBoxRow.__init__(self)
        self.option = option
        label = Gtk.Label(label=option.label)
        label.set_xalign(0)
        self.set_child(label)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk

from gajim.common import app
from gajim.common import types
from gajim.common.const import SimpleClientState
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager


class GroupchatState(Gtk.Box, SignalManager):
    def __init__(self) -> None:
        Gtk.Box.__init__(
            self, halign=Gtk.Align.CENTER, valign=Gtk.Align.END, visible=False
        )
        SignalManager.__init__(self)

        self._contact = None
        self._client = None

        self._ui = get_builder("groupchat_state.ui")
        self._connect(self._ui.join_button, "clicked", self._on_join_clicked)
        self._connect(self._ui.abort_join_button, "clicked", self._on_abort_clicked)
        self._connect(self._ui.close_button, "clicked", self._on_close_clicked)

        self.append(self._ui.groupchat_state)

    def do_unroot(self) -> None:
        self.clear()

        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def clear(self) -> None:
        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)

        if self._client is not None:
            self._client.disconnect_all_from_obj(self)

        self._contact = None
        self._client = None
        self.set_visible(False)

    def switch_contact(self, contact: types.ChatContactT) -> None:
        self.clear()

        if not isinstance(contact, GroupchatContact):
            return

        self._contact = contact
        self._contact.connect("state-changed", self._on_muc_state_changed)
        self._contact.connect("mam-sync-started", self._on_mam_sync_changed)
        self._contact.connect("mam-sync-finished", self._on_mam_sync_changed)
        self._contact.connect("mam-sync-error", self._on_mam_sync_error)

        self._client = app.get_client(contact.account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._update_state(contact)

    def _on_client_state_changed(
        self, client: types.Client, _signal_name: str, _state: SimpleClientState
    ) -> None:
        assert self._contact is not None
        self._update_state(self._contact)

    def _on_muc_state_changed(
        self, contact: GroupchatContact, _signal_name: str
    ) -> None:
        self._update_state(contact)

    def _update_state(self, contact: GroupchatContact) -> None:
        if contact.is_joining:
            self._ui.groupchat_state.set_visible_child_name("joining")

        elif contact.is_not_joined:
            self._ui.groupchat_state.set_visible_child_name("not-joined")

        assert self._client is not None
        self.set_visible(self._client.is_available() and not contact.is_joined)

    def _on_mam_sync_changed(
        self,
        _contact: GroupchatContact,
        signal_name: str,
    ) -> None:
        if signal_name == "mam-sync-started":
            self.set_visible(True)
            self._ui.groupchat_state.set_visible_child_name(signal_name)
            return

        self.set_visible(False)

    def _on_mam_sync_error(
        self, _contact: GroupchatContact, signal_name: str, error_text: str
    ) -> None:
        self.set_visible(True)
        self._ui.groupchat_state.set_visible_child_name(signal_name)
        self._ui.mam_error_label.set_text(
            _("There has been an error while trying to fetch messages: %s") % error_text
        )

    def _on_close_clicked(self, _button: Gtk.Button) -> None:
        self.set_visible(False)

    def _on_join_clicked(self, _button: Gtk.Button) -> None:
        assert self._contact is not None
        client = app.get_client(self._contact.account)
        client.get_module("MUC").join(self._contact.jid)

    def _on_abort_clicked(self, _button: Gtk.Button) -> None:
        assert self._contact is not None
        client = app.get_client(self._contact.account)
        client.get_module("MUC").abort_join(self._contact.jid)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast
from typing import Literal
from typing import overload

import logging

import pysequoia as pys
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.modules.openpgp import MultipleSecretKeysImportError
from gajim.common.modules.util import Task
from gajim.common.regex import OPENPGP_BACKUP_PASSWORD_RX
from gajim.common.util.text import format_fingerprint
from gajim.plugins.plugins_i18n import _

from gajim.gtk.activity_list import OpenPGPEvent
from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.openpgp.wizard")


class OpenPGPWizard(Assistant):
    def __init__(
        self, account: str, *, mode: Literal["test-password"] | None = None
    ) -> None:
        Assistant.__init__(
            self, name="OpenPGPWizard", height=500, transient_for=app.window, modal=True
        )
        self._destroyed: bool = False
        self._mode = mode

        self.account = account
        self._client = app.get_client(account)

        self.add_button("back", _("Back"))
        self.add_button("close", _("Close"))
        self.add_button("overwrite", _("Overwrite"), css_class="destructive-action")
        self.add_button("other-options", _("Other Options"))
        self.add_button(
            "import", _("Import Key"), complete=True, css_class="suggested-action"
        )
        self.add_button(
            "restore", _("Restore Key"), css_class="suggested-action", complete=True
        )
        self.add_button("choose", _("Confirm"), css_class="suggested-action")
        self.add_button(
            "test-password", _("Apply"), css_class="suggested-action", complete=True
        )

        self.add_pages(
            {
                "welcome": WelcomePage(),
                "import": ImportPage(),
                "password": BackupPasswordPage(mode=mode),
                "choose-secret-key": ChooseSecretKeyPage(),
                "error": ErrorPage(),
                "success": SuccessPage(),
            }
        )

        self.add_default_page("progress")

        welcome_page = self.get_page("welcome")
        self._connect(welcome_page, "clicked", self._on_welcome_page_button_clicked)

        self._connect(self, "button-clicked", self._on_assistant_button_clicked)
        self._connect(self, "page-changed", self._on_page_changed)

        self.show_page("progress")
        if (
            self._client.get_module("OpenPGP").secret_key_exists()
            and not mode == "test-password"
        ):
            self._show_success_page("key-exists")
        else:
            self._client.get_module("OpenPGP").request_secret_key(
                callback=self._on_secret_key_received
            )

    @overload
    def get_page(self, name: Literal["welcome"]) -> WelcomePage: ...

    @overload
    def get_page(self, name: Literal["import"]) -> ImportPage: ...

    @overload
    def get_page(self, name: Literal["password"]) -> BackupPasswordPage: ...

    @overload
    def get_page(self, name: Literal["choose-secret-key"]) -> ChooseSecretKeyPage: ...

    @overload
    def get_page(self, name: Literal["error"]) -> ErrorPage: ...

    @overload
    def get_page(self, name: Literal["success"]) -> SuccessPage: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def _on_assistant_button_clicked(
        self, _assistant: Assistant, button_name: str
    ) -> None:
        match button_name:
            case "import":
                key_str = self.get_page("import").get_key_string()
                try:
                    self._client.get_module("OpenPGP").import_key(key_str)
                except Exception as error:
                    self._show_error_page(
                        _("Import Error"), _("Import Error"), str(error)
                    )

            case "restore" | "choose":
                fingerprint = None
                if button_name == "choose":
                    fingerprint = self.get_page(
                        "choose-secret-key"
                    ).get_selected_fingerprint()

                password = self.get_page("password").get_password()

                try:
                    self._client.get_module("OpenPGP").import_key(
                        self._encrypted_backup_bytes, password, fingerprint
                    )
                except MultipleSecretKeysImportError as error:
                    self.get_page("choose-secret-key").set_certs(error.get_certs())
                    self.show_page("choose-secret-key")

                except Exception as error:
                    self._show_error_page(
                        _("Import Error"),
                        _("Import Error"),
                        _(
                            "An error occurred while trying to import your OpenPGP key: %s"
                        )
                        % str(error),
                    )

                else:
                    self._show_success_page("restored")

            case "test-password":
                password = self.get_page("password").get_password()
                try:
                    self._client.get_module("OpenPGP").test_backup_password(
                        self._encrypted_backup_bytes, password
                    )
                except Exception as error:
                    self._show_error_page(
                        _("Decryption Error"),
                        _("Decryption Error"),
                        _("An error occurred while trying to decrypt your backup: %s")
                        % str(error),
                    )

                else:
                    self._show_success_page("test-password")

            case "overwrite":
                self._client.get_module("OpenPGP").backup_secret_key(
                    callback=self._on_overwrite_result  # type: ignore
                )
                self.show_page("progress")

            case "other-options":
                self.show_page("welcome")

            case "back":
                self.show_last_page()

            case "close":
                self.close()

            case _:
                raise ValueError

    def _on_welcome_page_button_clicked(
        self, _page: Gtk.Widget, button_name: str
    ) -> None:
        if button_name == "import":
            self.show_page("import", Gtk.StackTransitionType.SLIDE_LEFT)

        elif button_name == "generate":
            try:
                self._client.get_module("OpenPGP").generate_key()
            except Exception as error:
                self._show_error_page(_("Error"), _("Error"), str(error))
            else:
                self._show_success_page("generate")

    def _on_page_changed(self, _assistant: Assistant, page_name: str) -> None:
        if page_name == "import":
            self.get_page("import").clear()

    def _show_success_page(
        self,
        mode: Literal[
            "generate", "overwrite", "key-exists", "restored", "test-password"
        ],
    ) -> None:

        password = None

        match mode:
            case "generate":
                title = _("Setup Complete")
                text = _(
                    "Your OpenPGP key has been created. Your chat is now encrypted."
                )
                password = self._client.get_module("OpenPGP").get_backup_password()

            case "overwrite" | "test-password":
                title = _("Backup Successful")
                text = _("Your OpenPGP key backup was successful.")
                if mode == "overwrite":
                    password = self._client.get_module("OpenPGP").get_backup_password()

            case "key-exists":
                title = _("Setup Complete")
                text = _("OpenPGP Setup is completed.")

            case "restored":
                title = _("Setup Complete")
                text = _("Successfully imported key from server.")

        success_page = self.get_page("success")
        success_page.set_title(title)
        success_page.set_text(text)
        success_page.show_backup_password(password)

        self.show_page("success", Gtk.StackTransitionType.SLIDE_LEFT)
        app.window.get_activity_list().remove_by_type(OpenPGPEvent, self.account)

    def _show_error_page(self, title: str, heading: str, text: str) -> None:
        self.get_page("error").set_title(title)
        self.get_page("error").set_heading(heading)
        self.get_page("error").set_text(text or "")
        self.show_page("error", Gtk.StackTransitionType.SLIDE_LEFT)

    def _on_secret_key_received(self, task: Task) -> None:
        try:
            encrypted_bytes = cast(bytes | None, task.finish())
        except Exception as error:
            log.error("Error on secret key request: %s", error)
            self.show_page("welcome", Gtk.StackTransitionType.SLIDE_LEFT)
            return

        if not encrypted_bytes:
            self.show_page("welcome", Gtk.StackTransitionType.SLIDE_LEFT)
            return

        self._encrypted_backup_bytes = encrypted_bytes
        self.show_page("password")

    def _on_overwrite_result(self, task: Task) -> None:
        try:
            task.finish()
        except Exception as error:
            log.error("Error when overwriting secret key : %s", error)
            self._show_error_page(
                _("Backup Error"),
                _("Backup Error"),
                _("An error occurred while trying to overwrite your backup: %s")
                % str(error),
            )

        else:
            self._show_success_page("overwrite")


@Gtk.Template(string=get_ui_string("openpgp/welcome.ui"))
class WelcomePage(AssistantPage):
    __gtype_name__ = "OpenPGPWizardWelcome"
    __gsignals__ = {
        "clicked": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    _import_button: Gtk.Button = Gtk.Template.Child()
    _generate_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Setup OpenPGP Encryption")

        self._connect(self._import_button, "clicked", self._on_button_clicked, "import")
        self._connect(
            self._generate_button, "clicked", self._on_button_clicked, "generate"
        )

    def _on_button_clicked(self, _button: Gtk.Button, name: str) -> None:
        self.emit("clicked", name)


@Gtk.Template(string=get_ui_string("openpgp/import.ui"))
class ImportPage(AssistantPage):
    __gtype_name__ = "OpenPGPWizardImport"

    _import_textview: Gtk.TextView = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.set_valign(Gtk.Align.FILL)

        self.title = _("Import OpenPGP Key")
        self.complete = False

        self._buffer = self._import_textview.get_buffer()
        self._connect(self._buffer, "changed", self._on_buffer_changed)

    def _on_buffer_changed(self, buffer: Gtk.TextBuffer) -> None:
        start, end = buffer.get_bounds()
        text = buffer.get_text(start, end, False)
        self.complete = bool(text)
        self.update_page_complete()

    def get_key_string(self) -> str:
        start, end = self._buffer.get_bounds()
        return self._buffer.get_text(start, end, False)

    def clear(self) -> None:
        self._buffer.delete(*self._buffer.get_bounds())

    def get_visible_buttons(self) -> list[str]:
        return ["back", "import"]


@Gtk.Template(string=get_ui_string("openpgp/password.ui"))
class BackupPasswordPage(AssistantPage):
    __gtype_name__ = "OpenPGPWizardBackupPassword"

    _password_entry: Gtk.PasswordEntry = Gtk.Template.Child()

    def __init__(self, mode: Literal["test-password"] | None) -> None:
        AssistantPage.__init__(self)
        self._mode = mode
        if mode == "test-password":
            self.title = _("Enter Password")
        else:
            self.title = _("Restore Backup")
        self.complete = False

        self._connect(self._password_entry, "changed", self._on_entry_changed)

    def get_visible_buttons(self) -> list[str]:
        if self._mode == "test-password":
            return ["overwrite", "test-password"]
        return ["other-options", "restore"]

    def _on_entry_changed(self, entry: Gtk.PasswordEntry) -> None:
        res = OPENPGP_BACKUP_PASSWORD_RX.fullmatch(self._password_entry.get_text())
        self.complete = res is not None
        self.update_page_complete()

    def get_password(self) -> str:
        return self._password_entry.get_text()


@Gtk.Template(string=get_ui_string("openpgp/choose_secret_key.ui"))
class ChooseSecretKeyPage(AssistantPage):
    __gtype_name__ = "OpenPGPWizardChooseSecretKey"

    _listbox: Gtk.ListBox = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Choose Secret Key")
        self.complete = False

        self._certs: list[pys.Cert] = []

        self._connect(self._listbox, "row-selected", self._on_row_selected)

    def _on_row_selected(
        self, listbox: Gtk.ListBox, row: Gtk.ListBoxRow | None
    ) -> None:
        self.complete = row is not None
        self.update_page_complete()

    def get_visible_buttons(self) -> list[str]:
        return ["back", "choose"]

    def get_selected_fingerprint(self) -> str:
        row = self._listbox.get_selected_row()
        assert row is not None
        label = cast(FingerprintLabel, row.get_child())
        return label.get_fingerprint()

    def set_certs(self, certs: list[pys.Cert]) -> None:
        container_remove_all(self._listbox)
        self._certs = certs
        for cert in certs:
            self._listbox.append(FingerprintLabel(cert))


@Gtk.Template(string=get_ui_string("openpgp/success.ui"))
class SuccessPage(AssistantPage):
    __gtype_name__ = "OpenPGPWizardSuccess"

    _heading: Gtk.Label = Gtk.Template.Child()
    _text: Gtk.Label = Gtk.Template.Child()
    _backup_password_box: Gtk.Box = Gtk.Template.Child()
    _password_entry: Gtk.PasswordEntry = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)

    def set_title(self, title: str) -> None:
        self.title = title

    def set_heading(self, heading: str) -> None:
        self._heading.set_text(heading)

    def set_text(self, text: str) -> None:
        self._text.set_text(text)

    def show_backup_password(self, backup_password: str | None) -> None:
        if backup_password is None:
            self._backup_password_box.set_visible(False)
            return

        self._backup_password_box.set_visible(True)
        self._password_entry.set_text(backup_password)

    def get_visible_buttons(self) -> list[str]:
        return ["close"]


class FingerprintLabel(Gtk.Label):
    def __init__(self, cert: pys.Cert) -> None:
        super().__init__()
        self._cert = cert
        self.set_label(format_fingerprint(cert.fingerprint, "OpenPGP", wrap=True))
        self.add_css_class("p-6")
        self.add_css_class("monospace")

    def get_fingerprint(self) -> str:
        return self._cert.fingerprint


class ErrorPage(AssistantErrorPage):
    def get_visible_buttons(self) -> list[str]:
        return ["back", "close"]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Generic
from typing import TypeVar

from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app

_K = TypeVar("_K")


class GajimDropDown(Gtk.DropDown, Generic[_K]):
    __gtype_name__ = "GajimDropDown"

    def __init__(
        self,
        data: dict[Any, str] | list[str] | None = None,
        fixed_width: int = -1,
    ) -> None:
        Gtk.DropDown.__init__(self)

        self._model: Gio.ListStore[KeyValueItem[_K]] = Gio.ListStore(
            item_type=KeyValueItem
        )
        list_store_expression = Gtk.PropertyExpression.new(
            KeyValueItem,
            None,
            "value",
        )

        self.set_expression(list_store_expression)
        self.set_search_match_mode(Gtk.StringFilterMatchMode.SUBSTRING)
        self.set_model(self._model)

        factory = Gtk.SignalListItemFactory()
        factory.connect("bind", self._on_factory_bind)

        self.set_factory(factory)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup, KeyValueViewListItem, {})
        factory.connect("bind", self._on_factory_bind)

        self.set_list_factory(factory)

        self._setup_signal_id = None
        self.fixed_width = fixed_width

        self.set_data(data)

    @GObject.Property(type=int, default=-1, flags=GObject.ParamFlags.READWRITE)
    def fixed_width(self) -> int:  # type: ignore
        return self._fixed_width

    @fixed_width.setter
    def fixed_width(self, value: int) -> None:
        self._fixed_width = value

        kwargs: dict[str, Any] = {
            "width_chars": value,
            "max_width_chars": value,
        }

        factory = self.get_factory()
        assert factory is not None

        if self._setup_signal_id is not None:
            factory.disconnect(self._setup_signal_id)
        self._setup_signal_id = factory.connect(
            "setup", self._on_factory_setup, KeyValueViewItem, kwargs
        )

    @staticmethod
    def _on_factory_setup(
        _factory: Gtk.SignalListItemFactory,
        list_item: Gtk.ListItem,
        view_item: Any,
        kwargs: dict[str, Any],
    ) -> None:
        list_item.set_child(view_item(**kwargs))

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = cast(KeyValueViewItem, list_item.get_child())
        obj = cast(KeyValueItem[_K], list_item.get_item())
        view_item.bind(obj)

    def set_data(self, data: dict[Any, str] | list[str] | None) -> None:
        self._model.remove_all()

        if not data:
            return

        items: list[KeyValueItem[_K]] = []
        if isinstance(data, dict):
            for key, value in data.items():
                items.append(KeyValueItem(key=key, value=value))

        if isinstance(data, list):
            for entry in data:
                items.append(KeyValueItem(key=entry, value=entry))  # type: ignore

        self._model.splice(0, 0, items)

    def get_selected_key(self) -> _K | None:
        selected_item = self.get_selected_item()
        if selected_item is None:
            return None
        return selected_item.key

    def get_selected_item(self) -> KeyValueItem[_K] | None:
        return cast(KeyValueItem[_K] | None, super().get_selected_item())

    def select_key(self, key: _K) -> None:
        for pos in range(self._model.get_n_items()):
            item = self._model.get_item(pos)
            assert item is not None
            if item.key == key:
                self.set_selected(pos)

    def select_first(self) -> None:
        if self._model.get_n_items() > 0:
            self.set_selected(0)

    def has_key(self, key: _K) -> bool:
        for pos in range(self._model.get_n_items()):
            item = self._model.get_item(pos)
            assert item is not None
            if item.key == key:
                return True
        return False

    def get_item_count(self) -> int:
        return self._model.get_n_items()

    def run_destroy(self) -> None:
        # item can carry tooltips, and GTK hides a visible tooltip only via gtk_widget_unroot().
        # Destroying the items while still rooted results in a leak.
        if self.get_root() is not None:
            self.unparent()

        self.set_model(None)
        self._model.remove_all()
        app.check_finalize(self._model)
        del self._model
        app.check_finalize(self)


class KeyValueItem(GObject.Object, Generic[_K]):
    key: _K = GObject.Property(type=object, flags=GObject.ParamFlags.READWRITE)  # type: ignore
    value: str = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)  # type: ignore

    def __init__(self, *, key: _K, value: str) -> None:
        GObject.Object.__init__(self, key=key, value=value)


class KeyValueViewItem(Gtk.Label):
    def __init__(self, **kwargs: Any):
        Gtk.Label.__init__(
            self, ellipsize=Pango.EllipsizeMode.MIDDLE, xalign=0, **kwargs
        )

    def bind(self, item: KeyValueItem[_K]) -> None:
        self.set_label(item.value)
        self.set_tooltip_text(item.value)


class KeyValueViewListItem(Gtk.Label):
    def __init__(self):
        Gtk.Label.__init__(
            self,
            xalign=0,
        )

    def bind(self, item: KeyValueItem[_K]) -> None:
        self.set_label(item.value)
        self.set_tooltip_text(item.value)

        if len(item.value) > 35:
            self.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
            self.set_width_chars(30)
        else:
            self.set_ellipsize(Pango.EllipsizeMode.NONE)
            self.set_width_chars(-1)
